%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1993, Ellemtel Telecommunications Systems Laboratories
%% File    : dbg_ieval.erl
%% Author  : Magnus Fr|berg
%% Purpose : An evaluator for Erlang abstract syntax.

-module(dbg_ieval).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

-export([
	 eval_fnk/10,
	 bif/2,
	 bif/3,
	 expr/6,
	 follow/4
	]).

-export([
	 format_args/1,
	 trace_out/1,
	 exit/4,
	 exit/5,
	 pop/1,
	 get_stack/0,
	 in_use_p/2
	]).

-import(dbg_imeta,[add_catch_lev/0,dec_catch_lev/0,get_catch_lev/0]).

%% FIXME: My understanding of the contents of a stack_frame
%% FIXME: Needs to be implemented throughout all modules.
-define(stack_frame,{level,
		     module,			%Current module
		     function,			%Current function, name/arity
		     line,			%Current line (source code?)
		     bindings			%Variable bindings
		     }).



%% seq(ExpressionSeq, Bindings, CurrentModule, LastCall, Level, Function)
%%  Returns:
%%	{value,Value,NewBindings}

seq([E], Bs, Cm, Lc, Le, F) when element(2,E) < 1 ->     % Lambda call
    expr(E, Bs, Cm, Lc, Le, F);
seq([E|Es], Bs0, Cm, Lc, Le, F) when element(2,E) < 1 -> % Lambda call
    {value,_,Bs} = expr(E, Bs0, Cm, false, Le, F),
    seq(Es, Bs, Cm, Lc, Le, F);
seq([E], Bs0, Cm, Lc, Le, F) ->
    case dbg_icmd:cmd(E,Bs0,Cm,Le,F) of
	{skip,Bs} ->
	    {value,skipped,Bs};
	Bs ->
	    expr(E, Bs, Cm, Lc, Le, F)
    end;
seq([E|Es], Bs0, Cm, Lc, Le, F) ->
    case dbg_icmd:cmd(E,Bs0,Cm,Le,F) of
	{skip,Bs} ->
	    seq(Es, Bs, Cm, Lc, Le, F);
	Bs1 ->
	    {value,_,Bs} = expr(E, Bs1, Cm, false, Le, F),
	    seq(Es, Bs, Cm, Lc, Le, F)
    end;
seq([], Bs, Cm, _, _, _) ->
    {value,true,Bs}.

%% expr(Expression, Bindings, CurrentModule, LastCall, Level, Function)
%%  Returns:
%%	{value,Value,NewBindings}

%% Atom
%% 
expr({atom,_,A}, Bs, _, _, _, _) ->
    {value,A,Bs};

%% Integer
%% 
expr({integer,_,I}, Bs, _, _, _, _) ->
    {value,I,Bs};

%% Variable
%%
expr({var,Line,V}, Bs, Cm, _, _, F) ->
    case binding(V, Bs) of
	{value,Val} ->
	    {value,Val,Bs};
	unbound ->
	    exit({{unbound,V},F},Cm,Line,Bs)
    end;

%% Tuple
%%
expr({tuple,Line,Es0}, Bs0, Cm, _, Le, F) ->
    {Vs,Bs} = eval_list(Es0, Bs0, Cm, Le, Line, F),
    {value,list_to_tuple(Vs),Bs};

%% Receive statement
%%
expr({'receive',Line,Cs}, Bs0, Cm, Lc, Le, F) ->
    trace_receive(Le,false),
    eval_receive(get(self),Cs,Bs0,Cm,Lc,Le,Line,F);

%% Float
%%
expr({float,_,F}, Bs, _, _, _, _) ->
    {value,F,Bs};

%% Character
%% 
expr({char,_,C}, Bs, _, _, _, _) ->
    {value,C,Bs};

%% String
%%
expr({string,_,S}, Bs, _, _, _, _) ->
    {value,S,Bs};

%% Catch statement
%%
expr({'catch',_,Expr}, Bs0, Cm, _, Le, F) ->
    add_catch_lev(),
    Msg_handler = get(self),
    case catch expr(Expr, Bs0, Cm, false, Le, F) of
	{value,Val,Bs} ->
	    dec_catch_lev(),
	    pop(Le),
	    {value,Val,Bs};
	{'EXIT',{Msg_handler,Reason}} ->
	    pop(Le),
	    exit({Msg_handler,Reason});
	Other ->
	    dec_catch_lev(),
	    pop(Le),
	    {value,Other,Bs0}
    end;

%% A block of statements
%%
expr({block,_,Es}, Bs, Cm, Lc, Le, F) ->
    seq(Es, Bs, Cm, Lc, Le, F);

%% If statement
%%
expr({'if',Line,Cs}, Bs, Cm, Lc, Le, F) ->
    if_clauses(Cs, Bs, Cm, Lc, Le, Line, F);

%% Fun
%%
expr({'fun',Line,{clauses,Cs}}, Bs, Cm, _, _, _) ->
    %% Only for fun's given from 'shell'.
    %% This is a really ugly hack (don't handle fun's
    %% with arity > 20) !
    case length(element(3,hd(Cs))) of
	0 -> {value,
	      fun () -> int_fun(Cs, [], Bs, Cm) end,
	      Bs};
	1 -> {value,
	      fun (A) -> int_fun(Cs, [A], Bs, Cm) end,
	      Bs};
	2 -> {value,
	      fun (A,B) -> int_fun(Cs, [A,B], Bs, Cm) end,
	      Bs};
	3 -> {value,
	      fun (A,B,C) -> int_fun(Cs, [A,B,C], Bs, Cm) end,
	      Bs};
	4 -> {value,
	      fun (A,B,C,D) -> int_fun(Cs, [A,B,C,D], Bs, Cm) end,
	      Bs};
	5 -> {value,
	      fun (A,B,C,D,E) -> int_fun(Cs, [A,B,C,D,E], Bs, Cm) end,
	      Bs};
	6 -> {value,
	      fun (A,B,C,D,E,F) -> int_fun(Cs, [A,B,C,D,E,F], Bs, Cm) end,
	      Bs};
	7 -> {value,
	      fun (A,B,C,D,E,F,G) -> int_fun(Cs, [A,B,C,D,E,F,G], Bs, Cm) end,
	      Bs};
	8 -> {value,
	      fun (A,B,C,D,E,F,G,H) -> int_fun(Cs, [A,B,C,D,E,F,G,H], Bs, Cm) end,
	      Bs};
	9 -> {value,
	      fun (A,B,C,D,E,F,G,H,I) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I], Bs, Cm) end,
	      Bs};
	10 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J], Bs, Cm) end,
	      Bs};
	11 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K], Bs, Cm) end,
	      Bs};
	12 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K], Bs, Cm) end,
	      Bs};
	13 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L,M) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K,L,M], Bs, Cm) end,
	      Bs};
	14 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L,M,N) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K,L,M,N], Bs, Cm) end,
	      Bs};
	15 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L,M,N,O) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O], Bs, Cm) end,
	      Bs};
	16 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P], Bs, Cm) end,
	      Bs};
	17 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q], Bs, Cm) end,
	      Bs};
	18 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R], Bs, Cm) end,
	      Bs};
	19 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S], Bs, Cm) end,
	      Bs};
	20 -> {value,
	      fun (A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T) -> int_fun(Cs, [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T], Bs, Cm) end,
	      Bs};
	Other ->
	    exit({'NYI',{'fun',Line,Cs}})
    end;

%% A cons (list)
%%
expr({cons,Line,H0,T0}, Bs0, Cm, _, Le, F) ->
    {value,H,Bs1} = expr(H0, Bs0, Cm, false, Le, F),
    {value,T,Bs2} = expr(T0, Bs0, Cm, false, Le, F),
    {value,[H|T],merge_bindings(Bs2, Bs1, Cm, Line, F)};

%% List comprehension expression
%% 
expr({lc,_,E,Qs}, Bs, Cm, Lc, Le, F) ->
    %% Only for lc's given from 'shell'.
    eval_lc(E, Qs, Bs, Cm, Lc, Le, F);

%% Matching expression (=)
%%
expr({match,Line,Lhs,Rhs0}, Bs0, Cm, Lc, Le, F) ->
    {value,Rhs,Bs1} = expr(Rhs0, Bs0, Cm, false, Le, F),
    case match(Lhs, Rhs, Bs1) of
	{match,Bs} ->
	    {value,Rhs,Bs};
	nomatch ->
	    exit({{badmatch,Rhs},F},Cm,Line,Bs1)
    end;

%% Case statement
%%
expr({'case',Line,E,Cs}, Bs0, Cm, Lc, Le, F) ->
    {value,Val,Bs} = expr(E, Bs0, Cm, false, Le, F),
    case_clauses(Val, Cs, Bs, Cm, Lc, Le, Line, F);

%% FIXME: Operators (?) Better description here
%%
expr({op,Line,Op,A0}, Bs0, Cm, Lc, Le, F) ->
    {value,A,Bs} = expr(A0, Bs0, Cm, false, Le, F),
    {value,eval_op(Op, A, Cm, Line, Bs, F),Bs};

%% Remote function call
%%
expr({call,Line,{remote,_,Mod,Func},As0}, Bs0, Cm, Lc, Le, F) ->
    {value,M,Bs1} = expr(Mod, Bs0, Cm, false, Le, F),
    {value,Fu,Bs2} = expr(Func, Bs0, Cm, false, Le, F),
    Bs = merge_bindings(Bs2, Bs1, Cm, Line, F),
    {As,Bs3} = eval_list(As0, Bs, Cm, Le, Line, F),
    case M of
	erlang ->  %% BIF !!!
	    eval_bif(Line,Fu,As,Bs3,Cm,Lc,Le,F);
	_ ->
	    trace_fnk_call(Le,Line,M,Fu,As),
	    eval_fnk(M, Fu, As, Bs3, Cm, Lc, Le, F, Line, extern)
    end;

%% Function Call
%%
expr({call,Line,Func0,As0}, Bs0, Cm, Lc, Le, F) ->
    {value,Func,Bs1} = expr(Func0, Bs0, Cm, false, Le, F),
    {As,Bs} = eval_list(As0, Bs1, Cm, Le, Line, F),
    if
	function(Func) ->
	    eval_bif(Line,apply,[Func,As],Bs,Cm,Lc,Le,F);
	true ->
	    case erl_internal:bif(Func, length(As)) of

		%% FIXME: The state the interpreter enters on a break
		%% appears to be different if waiting for a BIF than
		%% from an ordinary function.
		%% This causes problems for code-deletion. (And probably 
		%% other things as well).
		
		true ->
		    eval_bif(Line,Func,As,Bs,Cm,Lc,Le,F);
		false ->
		    trace_fnk_call(Le,Line,Cm,Func,As),
		    eval_fnk(Cm, Func, As, Bs, Cm, Lc, Le, F, Line, local)
	    end
    end;

%% Receive statement FIXME: What's the difference
%%
expr({'receive',Line,Cs,To,ToExprs}, Bs0, Cm, Lc, Le, F) ->
    {value,ToVal,ToBs} = expr(To,Bs0,Cm,false,Le,F),
    trace_receive(Le,true),
    check_timeoutvalue(ToVal,ToBs,To,Cm,F),
    {Stamp,_} = statistics(wall_clock),
    eval_receive(get(self),Cs,ToVal,ToExprs,ToBs,Bs0,Cm,Lc,Le,Line,0,Stamp,F);

%% Send (!)
%%
expr({op,Line,'!',T0,M0}, Bs0, Cm, Lc, Le, F) ->
    {value,T,Bs1} = expr(T0, Bs0, Cm, false, Le, F),
    {value,M,Bs2} = expr(M0, Bs0, Cm, false, Le, F),
    Bs = merge_bindings(Bs2, Bs1, Cm, Line, F),
    eval_send(Line, T, M, Bs, Cm, F);

%% In-line append (++)
%%
expr({op,Line,'++',L10,L20}, Bs0, Cm, Lc, Le, F) ->
    {value,L1,Bs1} = expr(L10, Bs0, Cm, false, Le, F),
    {value,L2,Bs2} = expr(L20, Bs0, Cm, false, Le, F),
    Bs = merge_bindings(Bs2, Bs1, Cm, Line, F),
    eval_bif(Line,append,[L1,L2],Bs,Cm,Lc,Le,F);

%% In-line subtract (--)
%%
expr({op,Line,'--',L10,L20}, Bs0, Cm, Lc, Le, F) ->
    {value,L1,Bs1} = expr(L10, Bs0, Cm, false, Le, F),
    {value,L2,Bs2} = expr(L20, Bs0, Cm, false, Le, F),
    Bs = merge_bindings(Bs2, Bs1, Cm, Line, F),
    eval_bif(Line,subtract,[L1,L2],Bs,Cm,Lc,Le,F);

%% Op Unknown? FIXME: What's the difference
%% 
expr({op,Line,Op,L0,R0}, Bs0, Cm, Lc, Le, F) ->
    {value,L,Bs1} = expr(L0, Bs0, Cm, false, Le, F),
    {value,R,Bs2} = expr(R0, Bs0, Cm, false, Le, F),
    Bs = merge_bindings(Bs2, Bs1, Cm, Line, F),
    {value,eval_op(Op, L, R, Cm, Line, Bs, F),Bs};

%% nil -  []
%%
expr({nil,_}, Bs, _, _, _, _) ->
    {value,[],Bs};

%% FIXME: Describe this
%%
expr({value,Val}, Bs, _, _, _, _) ->		%Special case straight values.
    {value,Val,Bs};

%% Records
%%
expr({record_index,_,Name,_}, _, _, _, _, _) ->
    exit({undef_record,Name});
expr({record,_,Name,_}, _, _, _, _, _) ->
    exit({undef_record,Name});
expr({record_field,_,_,Name,_}, _, _, _, _, _) ->
    exit({undef_record,Name});
expr({record,_,_,Name,_}, _, _, _, _, _) ->
    exit({undef_record,Name});
expr({record_field,_,_,_}, _, _, _, _, _) ->
    exit(undef_record);

%% FIXME: Describe this
%%
expr({remote,_,_,_}, _, _, _, _, F) ->
    exit({{badexpr,':'},in_fnk(F)});
expr({field,_,_,_}, _, _, _, _, F) ->
    exit({{badexpr,'.'},in_fnk(F)});

%% Brutal exit on unknown expressions/clauses/values/etc.
%%
expr(E, _, _, _, _, _) ->			%Not Yet Implemented
    exit({'NYI',E}).

%% eval_lc(Expr, [Qualifier], Bindings, Mod, LastCall, Level, Func) ->
%%	{value,Value,Bindings}.
%%  This is evaluating list comprehensions "straight out of the book".
%%  Copied from rv's implementation in erl_eval.

eval_lc(E, Qs, Bs, Cm, Lc, Le, F) ->
    {value,eval_lc1(E, Qs, Bs, Cm, Lc, Le, F), Bs}.

eval_lc1(E, [{generate,_,P,L0}|Qs], Bs0, Cm, Lc, Le, F) ->
    {value,L1,Bs1} = expr(L0, Bs0, Cm, false, Le, F),
    lists:flatmap(fun (V) ->
		    case match(P, V, []) of
			{match,Bsn} ->
			    Bs2 = add_bindings(Bsn, Bs1),
			    eval_lc1(E, Qs, Bs2, Cm, Lc, Le, F);
			nomatch -> []
		    end end, L1);
eval_lc1(E, [Q|Qs], Bs0, Cm, Lc, Le, F) ->
    case erl_lint:is_guard_test(Q) of
	true ->
	    case guard([Q], Bs0, Cm, Le) of
		true -> eval_lc1(E, Qs, Bs0, Cm, Lc, Le, F);
		false -> []
	    end;
	false ->
	    case expr(Q, Bs0, Cm, false, Le, F) of
		{value,true,Bs1} -> eval_lc1(E, Qs, Bs1, Cm, Lc, Le, F);
		{value,false,Bs1} -> [];
		Other -> exit({bad_filter,in_fnk(F)})
	    end
    end;
eval_lc1(E, [], Bs, Cm, Lc, Le, F) ->
    {value,V,_} = expr(E, Bs, Cm, false, Le, F),
    [V].

%% Entry point to an user defined (interactive in the attachment window
%% shell) fun.

int_fun(Cs, Args, Bs, Mod) ->
    int:eval(Mod, '$$int_eval$$', [Cs, Args, Bs]).

%% eval_fun(Clauses, Arguments, Bindings, Mod, LastCall, Level, Func) ->
%%	{value, Value, _Bs}

eval_fun([{clause,L,H,G,B}|Cs], As, Bs0, Cm, Lc, Le, F) ->
    case match_list(H, As, []) of
	{match,Bsn} ->				%The new bindings for the head
	    Bs1 = add_bindings(Bsn, Bs0),	% which then shadow!
	    case guard(G, Bs1, Cm, Le) of
		true ->
		    seq(B, Bs1, Cm, false, Le, F);
		false -> eval_fun(Cs, As, Bs0, Cm, Lc, Le, F)
	    end;
	nomatch ->
	    eval_fun(Cs, As, Bs0, Cm, Lc, Le, F)
    end;
eval_fun([], As, Bs, Cm, Lc, Le, F) ->
    exit({lambda_clause,in_fnk(F)}).

%% exit/4 -
%% Store the error reason and position if not
%% stored already.
%% Terminate according to type of exit.
%% -------------------------------------------------

exit({What,Where},Cm,Line,Bs) ->
    Reason = {What,in_fnk(Where)},
    put_error(Reason,Cm,Line,Bs),
    exit(Reason);
exit(Reason,Cm,Line,Bs) ->
    put_error(Reason,Cm,Line,Bs),
    exit(Reason).

confirmed_exit(Reason,Cm,Line,Bs) ->
    put_error(Reason,Cm,Line,Bs),
    exit({confirmed,Reason}).

%% exit/5

exit(Msg_handler,Reason,Cm,Line,Bs) ->  % We will only die if the real
    put_error(Reason,Cm,Line,Bs),       % process has terminated.
    exit({Msg_handler,Reason}).
    
throw(Value,Cm,Line,Bs) ->
    do_put_error(nocatch,Cm,Line,Bs),
    throw(Value).

put_error(Reason,Cm,Line,Bs) ->
    case get(error) of
	none when Line =/= -1 ->
	    do_put_error(Reason,Cm,Line,Bs);
	{R,_,_,_} when R =/= Reason, Line =/= -1 ->
	    do_put_error(Reason,Cm,Line,Bs);
	{Reason,{C,_},_,_} when C =/= Cm, Line =/= -1 ->
	    do_put_error(Reason,Cm,Line,Bs);
	{Reason,{Cm,L},_,_} when L =/= Line, Line =/= -1 ->
	    do_put_error(Reason,Cm,Line,Bs);
	_ ->
	    ok
    end.

do_put_error(Reason,Cm,Line,Bs) ->
    case get(guard) of
	true ->
	    ok;
	_ ->
	    BinStack = term_to_binary(get_stack()),
	    put(error,{Reason,{Cm,Line},Bs,BinStack}),
	    ok
    end.

%% -------------------------------------------------

%% Eval an user defined (interactive in attachment window) fun.
eval_fnk(Mod,'$$int_eval$$',[Cs,As,Bs],_,Cm,Lc,Le,F,Line,_) when integer(Le) ->
    push(Bs,F,Cm,Lc,Le,Line),
    case Lc == false of
	true ->
	    {value,Val,_} = eval_fun(Cs,As,Bs,Mod,true,Le+1,F),
	    trace_fnk_ret(Le,Val,false),
	    pop(Le),
	    {value,Val,Bs};
	_ ->
	    eval_fun(Cs,As,Bs,Mod,Lc,Le+1,F)
    end;
eval_fnk(Mod,Name,As,Bs,Cm,Lc,Le,F,Line,Called) when integer(Le) ->

    push(Bs,F,Cm,Lc,Le,Line),
    case dbg_imeta:function(Mod,Name,As,Called,Cm,Line,Bs) of

	%% Continue evaluation in non-interpreted code.
	{module,_} ->
	    Msg_handler = get(self),
	    Msg_handler ! {sys,self(),{catch_apply(get_catch_lev()),
				       Mod,Name,As}},
	    %% Enter meta-wait state
	    {value,Val,_} = wait_fnk(Msg_handler,Bs,Le,Cm,Line),
	    trace_fnk_ret(Le,Val,Lc),
	    pop(Le),
	    {value,Val,Bs};

	%% Continue evaluation in meta process.
	Cs when Lc == false ->
	    {value,Val,_} = fnk_clauses(Name,Cs,As,Mod,true,Le+1,1,Cm,Line,Bs),
	    trace_fnk_ret(Le,Val,Lc),
	    pop(Le),
	    {value,Val,Bs};
	Cs ->
	    %% Note, this MUST be the last call, since we are interpreting
	    %% code that executes their last call.
	    fnk_clauses(Name,Cs,As,Mod,Lc,Le+1,1,Cm,Line,Bs)
    end.

catch_apply(0) ->
    apply;
catch_apply(Lev) when integer(Lev) ->
    catch_apply.

%% wait_fnk/5 - Waits for a return value from the interpreted process.
%%
%% This function represents the state where the Meta process
%% is waiting for the interpreted process to evaluate non-interpreted
%% code.
%%
%% Returns:
%% {value, Value, Bindings}

wait_fnk(Msg_handler,Bs,Le,Cm,Line) when integer(Le) ->
    receive
	{sys,Msg_handler,{apply,Value}} ->
	    {value,Value,Bs};
	{sys,Msg_handler,{throwed,Value}} ->  % We have a catch !!
	    throw(Value);
	{sys,Msg_handler,{throwed_nocatch,Value}} ->
	    throw(Value,Cm,Line,Bs);
	{sys,Msg_handler,{exited_nocatch,Value}} ->
	    confirmed_exit(Value,Cm,Line,Bs);
	{error_handler,Msg_handler,{eval,Func}} ->
	    dbg_imeta:int(Msg_handler,Func,Le+1),
	    wait_fnk(Msg_handler,Bs,Le,Cm,Line);
	{attach,Msg_handler,AttPid} ->
	    dbg_icmd:attach(AttPid,Cm,Line),
	    AttPid ! {self(),func_at,Cm,Line,Le},
	    wait_fnk(Msg_handler,Bs,Le,Cm,Line);
	{break_msg,Type,Data} ->
	    dbg_icmd:break_msg(Type,Data),
	    wait_fnk(Msg_handler,Bs,Le,Cm,Line);

	%% Handle removal of interpreted code.
	{old_code, Module} ->
	    case in_use_p(Module,Cm) of
		%% A call to the Module is on the stack (or unknown), so we must terminate.
		true ->
%		    dbg_icmd:tell_attached({self(),old_code,Module}),
		    exit(get(self), kill),
		    dbg_ieval:exit(old_code, Cm, Line, Bs);
		
		false ->
		    erase([Module|db]),
		    erase(cache),
		    wait_fnk(Msg_handler,Bs,Le,Cm,Line)
	    end;


	%%FIXME: Handle new code
	{new_code, Module} ->
	    io:format("NYI: Handling of {new_code,Module} in: wait_fnk,Module:~p~n",[Module]),
	    wait_fnk(Msg_handler,Bs,Le,Cm,Line);

	Cmd when tuple(Cmd), element(1,Cmd) == cmd ->
	    dbg_icmd:wait_cmd(Cmd,Bs),
	    wait_fnk(Msg_handler,Bs,Le,Cm,Line);

	%% The interpreted process terminated
	{'EXIT',Msg_handler,Reason} ->
	    exit(Msg_handler,Reason,Cm,Line,Bs);

	
	%% Some other (?) linked process terminated. FIXME: Describe which
	{'EXIT',Pid,Reason} ->
	    case dbg_icmd:attached_p(Pid) of
		true ->
		    dbg_icmd:detach(Pid),
		    wait_fnk(Msg_handler,Bs,Le,Cm,Line);
		_ ->
		    exit(Reason,Cm,Line,Bs)
	    end
    end.

%% ---
%% Hold a stack which can be inspected, at break and exit.
%%  push(Bs,CurFnk,Cm,Lc,Le,Line),
%%
%% The push-function is conditional.
%% It determines whether or not to push stuff on the stack
%% by looking at the stack_trace option stored in the process 
%% dictonary
%% 
push(Bs,F,Cm,Lc,Le,Line) when Line > 0 ->
    push(Bs,F,Cm,Lc,Le,Line,get(stack_trace));
push(_,_,_,_,_,_) ->
    true.

push(Bs,F,Cm,_,Le,Line,all) ->
    put(stack,[{Le,{Cm,extract_fnk(F),Line,Bs}}|get_stack()]);
push(Bs,F,Cm,Lc,Le,Line,no_tail) when Lc == false ->
    put(stack,[{Le,{Cm,extract_fnk(F),Line,Bs}}|get_stack()]);
push(_,_,_,_,_,_,_) ->
    true.

extract_fnk({Cm,Fnk,Args,No}) -> {Fnk,length(Args)};
extract_fnk(_)                -> extern.

pop(Le) ->
    put(stack,pop(Le,get_stack())).

pop(Le,[{L,_}|Stack]) when Le < L ->
    pop(Le,Stack);
pop(Le,[{Le,_}|Stack]) ->
    Stack;
pop(_,Stack) ->
    Stack.

get_stack() ->
    case get(stack) of 
	undefined -> [];
	Stack     -> Stack
    end.

%% Returns true if Module is on the stack
in_use_p(Module, CurrentModule) when Module == CurrentModule ->
    true;

in_use_p(Module,_CurrentModule) ->
    StackStatus = get(stack_trace),
    case StackStatus of
	false ->
	    true;
	_Otherwise1 ->
	    Stack = get_stack(),
	    %% InFrameP returns 'true' if its argument is a stackframe that 
	    %% refers to the same module as the imported variable Module.
	    InFrameP =
		fun({Level, {CurrentModule,Function, Line, Bindings}}) ->
			case CurrentModule of
			    Module ->
				true;
			    _Otherwise2 ->
				false
			end
		end,
	    Result = lists:any(InFrameP, Stack),
	    Result
    end.


%% ---

eval_bif(Line,throw,[As],Bs,Cm,Lc,Le,F) ->
    Msg_handler = get(self),
    trace_bif_call(Le,Line,throw,[As]),
    case get_catch_lev() of
	0 ->  % No surrounding catch interpreted
	    Msg_handler ! {sys,self(),{catch_bif(get_catch_lev()),throw,[As]}},
	    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);
	_ ->
	    throw(As)
    end;
eval_bif(_,self,[],Bs,_,_,_,_) ->
    {value,get(self),Bs};
eval_bif(Line,apply,As,Bs,Cm,Lc,Le,F) ->
    Msg_handler = get(self),
    push(Bs,F,Cm,Lc,Le,Line),
    trace_bif_call(Le,Line,apply,As),
    Msg_handler ! {sys,self(),{catch_bif(get_catch_lev()),apply,As}},
    Res = wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F),
    pop(Le),
    Res;
eval_bif(Line,Name,As,Bs,Cm,Lc,Le,F) ->
    Msg_handler = get(self),
    trace_bif_call(Le,Line,Name,As),
    case safe_bif(Name) of
	false ->
	    Msg_handler ! {sys,self(),{catch_bif(get_catch_lev()),Name,As}},
	    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);

	spawn ->
	    Msg_handler ! {sys,self(),{catch_bif(get_catch_lev()),Name,As,
				       follow_mod()}},
	    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);
	_ ->
	    safe_bif(catch_bif(get_catch_lev()),Name,As,Bs,Cm,Lc,Le,Line,F)
    end.



%%
%% Ask attach-process where to start new attach-process
%% 

follow_mod() ->
    case get(state) of
	break ->
	    follow_mod1();
	_ ->
	    false
    end.

follow_mod1() ->
    case get(attached) of
	[AttP|_] ->
	    AttP ! {self(),which_func},
	    receive
		{AttP,this_func,{Mod,Fnk}} ->
		    {Mod,Fnk,AttP};
		{'EXIT',AttP,_} ->
		    dbg_icmd:detach(AttP),
		    follow_mod1()
	    end;
	_ ->
	    false
    end.

catch_bif(0) ->
    bif;
catch_bif(Lev) when integer(Lev) ->
    catch_bif.

safe_bif(Catch,Name,As,Bs,Cm,Lc,Le,Line,F) ->
    case catch bif(Name,As) of
	{'EXIT',Reason} ->
	    {Reason1, Reason2} = decode_bif_exit(Reason,F),
	    case Catch of
		bif ->
		    exit(Reason1,Cm,Line,Bs);
		catch_bif ->
		    exit(Reason2)
	    end;
	Value when Lc == true ->
	    {value,Value,Bs};
	Value ->
	    trace_fnk_ret(Le,Value,Lc),
	    {value,Value,Bs}
    end.

%% decode_bif_exit/2 -> {Reason1,Reason2}
%% where Reason2 is ready to drop at the top level,
%% i.e. the function # is stripped.
decode_bif_exit({'EXIT',{Reason,{?MODULE,bif,_}}},F) ->
    Res = {'EXIT',{Reason,in_fnk(F)}},
    {Res,Res};
decode_bif_exit({Reason,{?MODULE,bif,_}},F) ->
    {{Reason,F},{Reason,in_fnk(F)}};
decode_bif_exit(Reason,_) ->
    {Reason,Reason}.

%%
%% This function represents the state in which the Meta
%% process waits for the result of evaluating a BIF
%% in the interpreted process.
%%

wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F) ->
    receive
	{sys, Msg_handler, {bif, Val}} when Lc == false ->
	    trace_fnk_ret(Le,Val,Lc),
	    {value,Val,Bs};
	{sys, Msg_handler, {bif, Val}} ->
	    {value,Val,Bs};
	{sys, Msg_handler, {throwed, Value}} ->  % We have a catch !!
	    {_,Value1} = decode_bif_exit(Value,F),
	    throw(Value1);
	{sys,Msg_handler,{throwed_nocatch,Value}} ->
	    {_,Value1} = decode_bif_exit(Value,F),
	    throw(Value1,Cm,Line,Bs);
	{sys,Msg_handler,{exited_nocatch,Value}} ->
	    {_,Value1} = decode_bif_exit(Value,F),
	    confirmed_exit(Value1,Cm,Line,Bs);
	{error_handler,Msg_handler,{eval,Func}} ->   %% Only for apply
	    dbg_imeta:int(Msg_handler,Func,Le+1),
	    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);
	{attach,Msg_handler,AttPid} ->
	    dbg_icmd:attach(AttPid,Cm,Line),
	    AttPid ! {self(),func_at,Cm,Line,Le},
	    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);
	{break_msg,Type,Data} ->
	    dbg_icmd:break_msg(Type,Data),
	    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);

	%% Handle removal of interpreted code.
	{old_code, Module} ->
	    case in_use_p(Module,Cm) of
		%% A call to the Module is on the stack (or unknown), so we must terminate.
		true ->
%		    dbg_icmd:tell_attached({self(),old_code,Module}),
		    exit(get(self), kill),
		    dbg_ieval:exit(old_code, Cm, Line, Bs);
		false ->
		    erase([Module|db]),
		    erase(cache),
		    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F)
	    end;


	%%FIXME: Handle new code
	{new_code, Module} ->				
	    io:format("NYI: Handling of {new_code,Module} in: wait_bif,Module:~p~n",[Module]),
	    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);

	Cmd when tuple(Cmd), element(1,Cmd) == cmd ->
	    dbg_icmd:wait_cmd(Cmd,Bs),
	    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);
	{'EXIT', Msg_handler, Reason} ->
	    exit(Msg_handler,Reason,Cm,Line,Bs);
	{'EXIT',Pid,Reason} ->
	    case dbg_icmd:attached_p(Pid) of
		true ->
		    dbg_icmd:detach(Pid),
		    wait_bif(Msg_handler,Bs,Le,Lc,Cm,Line,F);
		_ ->
		    exit(Reason,Cm,Line,Bs)
	    end
    end.

eval_send(Line, To, Msg, Bs, Cm, F) ->
    case catch To ! Msg of
	Msg ->
	    trace_send(To,Msg),
	    {value,Msg,Bs};
	{'EXIT',{badarg,{T,M}}} ->  % Special if To is an non-existing name.
	    exit({badarg,{T,M}},Cm,Line,Bs);
	{'EXIT',Reason} ->          % Reason is badarg... if send fails !!
	    exit({badarg,F},Cm,Line,Bs);
	What ->
	    io:format('dbg_ieval:eval_send/5 - Unknown behaviour of send!\n'),
	    exit({badarg,F},Cm,Line,Bs)
    end.

%% Start tracing of messages before fetching current messages in the queue
%% to make sure that no messages are lost. 
	    
eval_receive(Msg_handler,Cs,Bs0,Cm,Lc,Le,Line,F) ->
    %% To avoid private message passing protocol between META
    %% and interpreted process.
    erlang:trace(Msg_handler,true,['receive']),

    {_,Msgs} = erlang:process_info(Msg_handler,messages),
%    io:format('Msgs = ~w~n',[Msgs]),
    case receive_clauses(Cs,Bs0,Cm,Le,Msgs) of
	nomatch ->
	    dbg_iserver_api:set_state(waiting),
	    tell_att_if_break(wait_at,Cm,Line,Le),
	    eval_receive1(Msg_handler,Cs,Bs0,Cm,Lc,Le,Line,F);
	{eval,B,Bs,Msg} ->
	    rec_mess(Msg_handler,Msg,Cm,Line,Bs0),
	    seq(B,Bs,Cm,Lc,Le,F)
    end.

eval_receive1(Msg_handler,Cs,Bs0,Cm,Lc,Le,Line,F) ->
    Msgs = do_receive(Msg_handler,Cm,Le,Line,Bs0),
%    io:format('Msgs = ~w~n',[Msgs]),
    case receive_clauses(Cs,Bs0,Cm,Le,Msgs) of
	nomatch ->
	    eval_receive1(Msg_handler,Cs,Bs0,Cm,Lc,Le,Line,F);
	{eval,B,Bs,Msg} ->
	    rec_mess(Msg_handler,Msg,Cm,Line,Bs0),
	    dbg_iserver_api:set_state(running),
	    tell_att_if_break(running,Cm,Line,Le),
	    seq(B,Bs,Cm,Lc,Le,F)
    end.

check_timeoutvalue(ToVal,_,_,_,_) when integer(ToVal), ToVal >= 0 -> true;
check_timeoutvalue(infinity,_,_,_,_)                              -> true;
check_timeoutvalue(ToVal,ToBs,To,Cm,F) ->
    Line = element(2,To),
    exit({timeout_value,F},Cm,Line,ToBs).

eval_receive(Msg_handler,Cs,0,ToExprs,ToBs,Bs0,Cm,Lc,Le,Line,0,Stamp,F) -> % Timeout after 0ms
    {_,Msgs} = erlang:process_info(Msg_handler,messages),                  % at the first call
%    io:format('Msgs = ~w~n',[Msgs]),
    case receive_clauses(Cs,Bs0,Cm,Le,Msgs) of
	{eval,B,Bs,Msg} ->
	    rec_mess_no_trace(Msg_handler,Msg,Cm,Line,Bs0),
	    seq(B,Bs,Cm,Lc,Le,F);
	nomatch ->
	    trace_received(),
	    seq(ToExprs,ToBs,Cm,Lc,Le,F)
    end;
eval_receive(Msg_handler,Cs,ToVal,ToExprs,ToBs,Bs0,Cm,Lc,Le,Line,0,Stamp,F) -> % Called first time
    erlang:trace(Msg_handler,true,['receive']),
    {_,Msgs} = erlang:process_info(Msg_handler,messages),
%    io:format('Msgs = ~w~n',[Msgs]),
    case receive_clauses(Cs,Bs0,Cm,Le,Msgs) of
	nomatch ->
	    {Stamp1,Time1} = newtime(Stamp,ToVal),
	    dbg_iserver_api:set_state(waiting),
	    tell_att_if_break(wait_after_at,Cm,Line,Le),
	    eval_receive(Msg_handler,Cs,Time1,ToExprs,ToBs,Bs0,Cm,Lc,Le,
			 Line,infinity,Stamp1,F);
	{eval,B,Bs,Msg} ->
	    rec_mess(Msg_handler,Msg,Cm,Line,Bs0),
	    seq(B,Bs,Cm,Lc,Le,F)
    end;
eval_receive(Msg_handler,Cs,ToVal,ToExprs,ToBs,Bs0,Cm,Lc,Le,Line,_,Stamp,F) -> % Now wait for 
    case do_receive(Msg_handler,ToVal,Stamp,Cm,Le,Line,Bs0) of % real timeout
	timeout ->
	    trace_received(),
	    rec_mess(Msg_handler),
	    dbg_iserver_api:set_state(running),
	    tell_att_if_break(running,Cm,Line,Le),
	    seq(ToExprs,ToBs,Cm,Lc,Le,F);
	Msgs ->
%            io:format('Msgs = ~w~n',[Msgs]),
	    case receive_clauses(Cs,Bs0,Cm,Le,Msgs) of
		nomatch ->
		    {Stamp1,Time1} = newtime(Stamp,ToVal),
		    eval_receive(Msg_handler,Cs,Time1,ToExprs,ToBs,Bs0,Cm,Lc,Le,Line,
				 infinity,Stamp1,F);
		{eval,B,Bs,Msg} ->
		    rec_mess(Msg_handler,Msg,Cm,Line,Bs0),
		    dbg_iserver_api:set_state(running),
		    tell_att_if_break(running,Cm,Line,Le),
		    seq(B,Bs,Cm,Lc,Le,F)
	    end
    end.

tell_att_if_break(running,_,_,_) ->
    case get(state) of
	break -> dbg_icmd:tell_attached({self(),running});
	_     -> ok
    end;
tell_att_if_break(RecAt,Cm,Line,Le) ->
    case get(state) of
	break -> dbg_icmd:tell_attached({self(),RecAt,Cm,Line,Le});
	_     -> ok
    end.

do_receive(Msg_handler,Cm,Le,Line,Bs) ->
    receive
	{trace,'receive',Msg_handler,Msg} ->
	    [Msg];
	{cmd,Cmd,Args} ->
	    dbg_icmd:cmd_rec(Cmd, Args, Bs),
	    do_receive(Msg_handler,Cm,Le,Line,Bs);
	{break_msg,Type,Data} ->
	    dbg_icmd:break_msg(Type,Data),
	    do_receive(Msg_handler,Cm,Le,Line,Bs);
	{attach,Msg_handler,AttPid} ->
	    dbg_icmd:attach(AttPid,Cm,Line),
	    AttPid ! {self(),wait_at,Cm,Line,Le},
	    do_receive(Msg_handler,Cm,Le,Line,Bs);

	%% Handle removal of interpreted code.
	{old_code, Module} ->
	    case in_use_p(Module, Cm) of
		%% A call to the Module is on the stack (or unknown), so we must terminate.
		true ->
%		    dbg_icmd:tell_attached({self(),old_code,Module}),
		    exit(get(self), kill),
		    dbg_ieval:exit(old_code, Cm, Line, Bs);
		false ->
		    erase([Module|db]),
		    erase(cache),
		    do_receive(Msg_handler,Cm,Le,Line,Bs)
	    end;


	%%FIXME: Handle new_code.
	{new_code, Module} ->
	    io:format("NYI: Handling of {new_code,Module} in: do_receive,Module:~p~n",[Module]),
	    do_receive(Msg_handler,Cm,Le,Line,Bs);



	{'EXIT',Msg_handler,Reason} ->
	    exit(Msg_handler,Reason,Cm,Line,Bs);
	{'EXIT',Pid,Reason} ->
	    case dbg_icmd:attached_p(Pid) of
		true ->
		    dbg_icmd:detach(Pid),
		    do_receive(Msg_handler,Cm,Le,Line,Bs);
		_ ->
		    exit(Reason,Cm,Line,Bs)
	    end
    end.

do_receive(Msg_handler,Time,Stamp,Cm,Le,Line,Bs) ->
    receive
	{trace,'receive',Msg_handler,Msg} ->
	    [Msg];
	{cmd,timeout,_} ->
	    timeout;
	{cmd,Cmd,Args} ->
	    dbg_icmd:cmd_rec(Cmd, Args, Bs),
	    {Stamp1,Time1} = newtime(Stamp,Time),
	    do_receive(Msg_handler,Time1,Stamp1,Cm,Le,Line,Bs);
	{break_msg,Type,Data} ->
	    dbg_icmd:break_msg(Type,Data),
	    {Stamp1,Time1} = newtime(Stamp,Time),
	    do_receive(Msg_handler,Time1,Stamp1,Cm,Le,Line,Bs);
	{attach,Msg_handler,AttPid} ->
	    dbg_icmd:attach(AttPid,Cm,Line),
	    AttPid ! {self(),wait_after_at,Cm,Line,Le},
	    {Stamp1,Time1} = newtime(Stamp,Time),
	    do_receive(Msg_handler,Time1,Stamp1,Cm,Le,Line,Bs);

		%% Handle removal of interpreted code.
	{old_code, Module} ->
	    case in_use_p(Module, Cm) of
		%% A call to the Module is on the stack (or unknown), so we must terminate.
		true ->
%		    dbg_icmd:tell_attached({self(),old_code,Module}),
		    exit(get(self), kill),
		    dbg_ieval:exit(old_code, Cm, Line, Bs);
		false ->
		    erase([Module|db]),
		    erase(cache),
		    do_receive(Msg_handler,Time,Stamp,Cm,Le,Line,Bs)
	    end;

	%%FIXME: Handle new_code.
	{new_code, Module} ->
	    io:format("NYI: Handling of {new_code,Module} in: do_receive,Module:~p~n",[Module]),	    
	    do_receive(Msg_handler,Time,Stamp,Cm,Le,Line,Bs);

	{'EXIT',Msg_handler,Reason} ->
	    exit(Msg_handler,Reason,Cm,Line,Bs);
	{'EXIT',Pid,Reason} ->
	    case dbg_icmd:attached_p(Pid) of
		true ->
		    dbg_icmd:detach(Pid),
		    {Stamp1,Time1} = newtime(Stamp,Time),
		    do_receive(Msg_handler,Time1,Stamp1,Cm,Le,Line,Bs);
		_ ->
		    exit(Reason,Cm,Line,Bs)
	    end
    after Time ->
	    timeout
    end.

newtime(Stamp,infinity) ->
    {Stamp,infinity};
newtime(Stamp,Time) ->
    {Stamp1,_} = statistics(wall_clock),
    case Time - (Stamp1 - Stamp) of
	NewTime when NewTime > 0 ->
	    {Stamp1,NewTime};
	_ ->
	    {Stamp1,0}
    end.

rec_mess(Msg_handler,Msg,Cm,Line,Bs) ->
    erlang:trace(Msg_handler,false,['receive']),
    flush_traces(Msg_handler),
    Msg_handler ! {sys,self(),{'receive',Msg}},
    rec_ack(Msg_handler,Cm,Line,Bs).

rec_mess(Msg_handler) ->
    erlang:trace(Msg_handler,false,['receive']),
    flush_traces(Msg_handler).

rec_mess_no_trace(Msg_handler,Msg,Cm,Line,Bs) ->
    Msg_handler ! {sys,self(),{'receive',Msg}},
    rec_ack(Msg_handler,Cm,Line,Bs).

rec_ack(Msg_handler,Cm,Line,Bs) ->
    receive
	{Msg_handler,rec_acked} ->
	    true;
	{'EXIT',Msg_handler,Reason} ->
	    exit(Msg_handler,Reason,Cm,Line,Bs)
    end.

flush_traces(Msg_handler) ->
    receive
	{trace,'receive',Msg_handler,_} ->
	    flush_traces(Msg_handler)
    after 0 ->
	    true
    end.

%% eval_list(ExpressionList, Bindings, CurrentModule, Level, Line, Function)
%%  Evaluate a list of expressions "in parallel" at the same level.

eval_list(Es, Bs, Cm, Le, Line, F) ->
    eval_list(Es, [], Bs, Bs, Cm, Le, Line, F).

eval_list([E|Es], Vs, BsOrig, Bs0, Cm, Le, Line, F) ->
    {value,V,Bs1} = expr(E, BsOrig, Cm, false, Le, F),
    eval_list(Es, [V|Vs], BsOrig, merge_bindings(Bs1, Bs0, Cm, Line, F),
	      Cm, Le, Line, F);
eval_list([], Vs, _, Bs, _, _, _, _) ->
    {reverse(Vs,[]),Bs}.

eval_op('+', A1, A2, Cm, Line, Bs, _) when number(A1), number(A2) ->
    A1 + A2;
eval_op('-', A1, A2, Cm, Line, Bs, _) when number(A1), number(A2) ->
    A1 - A2;
eval_op('*', A1, A2, Cm, Line, Bs, _) when number(A1), number(A2) ->
    A1 * A2;
eval_op('/', A1, A2, Cm, Line, Bs, _) when number(A1), number(A2) ->
    A1 / A2;
eval_op('div', A1, A2, Cm, Line, Bs, _) when integer(A1), integer(A2) ->
    A1 div A2;
eval_op('rem', A1, A2, Cm, Line, Bs, _) when integer(A1), integer(A2) ->
    A1 rem A2;
eval_op('band', A1, A2, Cm, Line, Bs, _) when integer(A1), integer(A2) ->
    A1 band A2;
eval_op('bor', A1, A2, Cm, Line, Bs, _) when integer(A1), integer(A2) ->
    A1 bor A2;
eval_op('bxor', A1, A2, Cm, Line, Bs, _) when integer(A1), integer(A2) ->
    A1 bxor A2;
eval_op('bsl', A1, A2, Cm, Line, Bs, _) when integer(A1), integer(A2) ->
    A1 bsl A2;
eval_op('bsr', A1, A2, Cm, Line, Bs, _) when integer(A1), integer(A2) ->
    A1 bsr A2;
eval_op('^', A1, A2, Cm, Line, Bs, _) when number(A1), number(A2) ->
    math:pow(A1, A2);
%% These can be trimmed when the operations are implemented in the emulator.
eval_op('<', E1, E2, _, _, _, _) when E1 < E2 -> true;
eval_op('<', E1, E2, _, _, _, _) -> false;
eval_op('=<', E1, E2, _, _, _, _) when E1 =< E2 -> true;
eval_op('=<', E1, E2, _, _, _, _) -> false;
eval_op('>', E1, E2, _, _, _, _) when E1 > E2 -> true;
eval_op('>', E1, E2, _, _, _, _) -> false;
eval_op('>=', E1, E2, _, _, _, _) when E1 >= E2 -> true;
eval_op('>=', E1, E2, _, _, _, _) -> false;
eval_op('==', E1, E2, _, _, _, _) when E1 == E2 -> true;
eval_op('==', E1, E2, _, _, _, _) -> false;
eval_op('/=', E1, E2, _, _, _, _) when E1 /= E2 -> true;
eval_op('/=', E1, E2, _, _, _, _) -> false;
eval_op('=:=', E1, E2, _, _, _, _) when E1 =:= E2 -> true;
eval_op('=:=', E1, E2, _, _, _, _) -> false;
eval_op('=/=', E1, E2, _, _, _, _) when E1 =/= E2 -> true;
eval_op('=/=', E1, E2, _, _, _, _) -> false;
eval_op('and', E1, E2, _, _, _, _) -> E1 and E2;
eval_op('or', E1, E2, _, _, _, _) -> E1 or E2;
eval_op('xor', E1, E2, _, _, _, _) -> E1 xor E2;
eval_op(Op, A1, A2, Cm, Line, Bs, F) ->
    exit({badarith,F},Cm,Line,Bs).

eval_op('+', A, Cm, Line, Bs, _) when number(A) ->
    A;
eval_op('-', A, Cm, Line, Bs, _) when number(A) ->
    -A;
eval_op('bnot', A, Cm, Line, Bs, _) when integer(A) ->
    bnot A;
eval_op('not', A, Cm, Line, Bs, _) ->
    not A;
eval_op(Op, A, Cm, Line, Bs, F) ->
    exit({badarith,F},Cm,Line,Bs).

%% bif(Name, Arguments)
%%  Evaluate the Erlang builtin function Name. (Special case apply here)

bif(apply, [M,F,As]) -> 
    apply(M, F, As);
bif(apply, [F,As]) -> 
    apply(F, As);
bif(Name, As) ->
    apply(erlang, Name, As).

bif(Name, As, false) ->
    bif(Name, As);
bif(spawn, [M,F,As], Attached) ->
    spawn(?MODULE,follow,[Attached,M,F,As]);
bif(spawn_link, [M,F,As], Attached) ->
    spawn_link(?MODULE,follow,[Attached,M,F,As]);
bif(spawn, [N,M,F,As], Attached) ->
    spawn(N,?MODULE,follow,[Attached,M,F,As]);
bif(spawn_link, [N,M,F,As], Attached) ->
    spawn_link(N,?MODULE,follow,[Attached,M,F,As]).

follow(Fol,M,F,As) ->
    dbg_iserver_api:am_followed(Fol),
    apply(M,F,As).

%% if_clauses(Clauses, Bindings, CurrentModule, LastCall, Line, Function)

if_clauses([{clause,_,[],G,B}|Cs], Bs, Cm, Lc, Le, Line, F) ->
    case guard(G, Bs, Cm, Le) of
	true ->
	    seq(B, Bs, Cm, Lc, Le, F);
	false ->
	    if_clauses(Cs, Bs, Cm, Lc, Le, Line, F)
    end;
if_clauses([], Bs, Cm, Lc, Le, Line, F) ->
    exit({if_clause,F},Cm,Line,Bs).

%% case_clauses(Value, Clauses, Bindings, CurrentModule, LastCall, Line, Function)

case_clauses(Val, [{clause,_,[P],G,B}|Cs], Bs0, Cm, Lc, Le, Line, F) ->
    case match(P, Val, Bs0) of
	{match,Bs} ->
	    case guard(G, Bs, Cm, Le) of
		true ->
		    seq(B, Bs, Cm, Lc, Le, F);
		false ->
		    case_clauses(Val, Cs, Bs0, Cm, Lc, Le, Line, F)
	    end;
	nomatch ->
	    case_clauses(Val, Cs, Bs0, Cm, Lc, Le, Line, F)
    end;
case_clauses(Val, [], Bs, Cm, Lc, Le, Line, F) ->
    exit({{case_clause,Val},F},Cm,Line,Bs).

fnk_clauses(F,[{clause,_,Pars,G,B}|Cs],Args,Cm,Lc,Le,Clause,CallM,Line,Bs0)
  when integer(Clause) ->
    case fnk_match(F,Pars,Args,[]) of
	{match,Bs} ->
	    case guard(G,Bs,Cm,Le) of
		true ->
%		    io:format('fnk_clause ~w matched~n',[Clause]),
		    seq(B,Bs,Cm,Lc,Le,{Cm,F,Args,fnk_no()});
		false ->
		    fnk_clauses(F,Cs,Args,Cm,Lc,Le,Clause+1,CallM,Line,Bs0)
	    end;
	nomatch ->
	    fnk_clauses(F,Cs,Args,Cm,Lc,Le,Clause+1,CallM,Line,Bs0)
    end;
fnk_clauses(F,[],Args,Cm,Lc,Le,Clause,CallM,Line,Bs0) ->
    exit({function_clause,{Cm,F,Args,fnk_no()}},CallM,Line,Bs0).

%% Identifies each funktion call uniquely.
%% Counts the number of interpreted function calls.
%% Used especially for the finish functionality.
fnk_no() ->
    Number = get(fnk_no) + 1,
    put(fnk_no,Number),
    Number.

%Strip fnk_no !
in_fnk({M,F,A,No}) when integer(No) -> {M,F,A};
in_fnk(Fnk)                         -> Fnk.

receive_clauses(Cs,Bs0,Cm,Le,[Msg|Msgs]) ->
    case rec_clauses(Cs,Bs0,Cm,Le,Msg,1) of
	nomatch ->
	    receive_clauses(Cs,Bs0,Cm,Le,Msgs);
	{eval,B,Bs} ->
	    {eval,B,Bs,Msg}
    end;
receive_clauses(_,_,_,_,[]) ->
    nomatch.

rec_clauses([{clause,_,Pars,G,B}|Cs],Bs0,Cm,Le,Msg,Clause) 
  when integer(Clause) ->
    case rec_match(Pars,Msg,Bs0) of
	{match,Bs} ->
	    case guard(G,Bs,Cm,Le) of
		true ->
%		    io:format('receive_clause ~w matched~n',[Clause]),
		    trace_received(Msg),
		    {eval,B,Bs};
		false ->
		    rec_clauses(Cs,Bs0,Cm,Le,Msg,Clause+1)
	    end;
	nomatch ->
	    rec_clauses(Cs,Bs0,Cm,Le,Msg,Clause+1)
    end;
rec_clauses([],_,_,_,_,_) ->
    nomatch.


%% guard(GuardTests, Bindings)
%% The guard can never crash, due to compiler restrictions
%% for guard tests. The function pointer is therefore omitted.
%% Simulate a catch for evaluation of exprs.

guard([{call,Line,{atom,_,Name},As0}|Gs], Bs0, Cm, Le) ->
    add_catch_lev(),
    put(guard,true),
    case catch eval_list(As0, Bs0, Cm, Le, Line, {}) of
	{'EXIT',_} ->
	    dec_catch_lev(),
	    put(guard,false),
	    false;
	{As,Bs} ->
	    dec_catch_lev(),
	    put(guard,false),
	    case erl_internal:type_test(Name, length(As)) of
		true ->
		    case guard_test(Name, As) of
			true ->
			    guard(Gs, Bs, Cm, Le);
			false ->
			    false
		    end;
		false ->
		    false
	    end
    end;
guard([{op,_,Op,Lhs0,Rhs0}|Gs], Bs0, Cm, Le) ->
    add_catch_lev(),
    put(guard,true),
    case catch guard_comp_exprs(Lhs0, Rhs0, Bs0, Cm, Le) of
	{'EXIT',_} ->
	    dec_catch_lev(),
	    put(guard,false),
	    false;
	{Lhs,Rhs,Bs} ->
	    dec_catch_lev(),
	    put(guard,false),
	    case erl_internal:comp_op(Op, 2) of
		true ->
		    case guard_comp(Op, Lhs, Rhs) of
			true ->
			    guard(Gs, Bs, Cm, Le);
			false ->
			    false
		    end;
		false ->
		    false
	    end
    end;
guard([{atom,_,true}|Gs], Bs, Cm, Le) ->
    guard(Gs, Bs, Cm, Le);
guard([{atom,_,false}|Gs], _, _, _) ->
    false;
guard([], _, _, _) ->
    true.

guard_comp_exprs(Lhs0, Rhs0, Bs0, Cm, Le) ->
    {value,Lhs,Bs1} = expr(Lhs0, Bs0, Cm, false, Le, {}),
    {value,Rhs,Bs} = expr(Rhs0, Bs1, Cm, false, Le, {}),
    {Lhs,Rhs,Bs}.

guard_test(integer, [A]) when integer(A) -> true;
guard_test(float, [A]) when float(A) -> true;
guard_test(number, [A]) when number(A) -> true;
guard_test(atom, [A]) when atom(A) -> true;
guard_test(constant, [A]) when constant(A) -> true;
guard_test(list, [A]) when list(A) -> true;
guard_test(tuple, [A]) when tuple(A) -> true;
guard_test(pid, [A]) when pid(A) -> true;
guard_test(reference, [A]) when reference(A) -> true;
guard_test(port, [A]) when port(A) -> true;
guard_test(binary, [A]) when binary(A) -> true;
guard_test(function, [A]) when function(A) -> true;
guard_test(_, _) ->
    false.

%% Arithmetic comparison operators
guard_comp('==', Lhs, Rhs) when Lhs == Rhs -> true;
guard_comp('/=', Lhs, Rhs) when Lhs /= Rhs -> true;
guard_comp('=<', Lhs, Rhs) when Lhs =< Rhs -> true;
guard_comp('<', Lhs, Rhs) when Lhs < Rhs -> true;
guard_comp('>=', Lhs, Rhs) when Lhs >= Rhs -> true;
guard_comp('>', Lhs, Rhs) when Lhs > Rhs -> true;
%% Term comparison operators
guard_comp('=:=', Lhs, Rhs) when Lhs =:= Rhs -> true;
guard_comp('=/=', Lhs, Rhs) when Lhs =/= Rhs -> true;
guard_comp(_, _, _) -> false.

%% match(Pattern, Term, Bindings)
%%  Try to match Pattern against Term with the current bindings.
%%  Returns:
%%	{match,NewBindings}
%%	nomatch

match(Pat, Term, Bs) ->
    catch match1(Pat, Term, Bs).

match1({atom,_,A}, A, Bs) ->
    {match,Bs};
match1({integer,_,I}, I, Bs) ->
    {match,Bs};
match1({float,_,F}, F, Bs) ->
    {match,Bs};
match1({char,_,C}, C, Bs) ->
    {match,Bs};
match1({var,_,'_'}, Term, Bs) ->		%Anonymous variable matches
    {match,add_anon(Term,Bs)};			% everything, save it anyway.
match1({var,_,Name}, Term, Bs) ->
    case binding(Name, Bs) of
	{value,Term} ->
	    {match,Bs};
	{value,V} ->
	    throw(nomatch);
	unbound ->
	    {match,[{Name,Term}|Bs]}            %Add the new binding
    end;
match1({string,_,S}, S, Bs) ->
    {match,Bs};
match1({nil,_}, [], Bs) ->
    {match,Bs};
match1({cons,_,H,T}, [H1|T1], Bs0) ->
    {match,Bs} = match1(H, H1, Bs0),
    match1(T, T1, Bs);
match1({tuple,_,Elts}, Tuple, Bs) when length(Elts) == size(Tuple) ->
    match_tuple(Elts, Tuple, 1, Bs);
match1({op,_,'-',A}, Term, Bs) when number(Term) ->
    match1(A, -Term, Bs);
match1({op,_,'+',A}, Term, Bs) when number(Term) ->
    match1(A, Term, Bs);
match1(_, _, _) ->
    throw(nomatch).

match_tuple([E|Es], Tuple, I, Bs0) when integer(I) ->
    {match,Bs} = match1(E, element(I, Tuple), Bs0),
    match_tuple(Es, Tuple, I+1, Bs);
match_tuple([], _, _, Bs) ->
    {match,Bs}.

%% match_list(PatternList, TermList, Bindings) ->
%%	{match,NewBindings} | nomatch
%%  Try to match a list of patterns against a list of terms with the
%%  current bindings.

match_list(Ps, Ts, Bs) ->
    catch match_list1(Ps, Ts, Bs).

match_list1([P|Ps], [T|Ts], Bs0) ->
    case match(P, T, Bs0) of
	{match,Bs1} -> match_list1(Ps, Ts, Bs1);
	nomatch -> throw(nomatch)
    end;
match_list1([], [], Bs) ->
    {match,Bs}.

fnk_match(F,[Par|Pars],[Arg|Args],Bs0) ->
    case match(Par, Arg, Bs0) of
	{match,Bs} ->
	    fnk_match(F,Pars,Args,Bs);
	nomatch ->
	    nomatch
    end;
fnk_match(F,[],[],Bs) ->
    {match,Bs}.

rec_match([Par],Msg,Bs0) ->
    match(Par,Msg,Bs0).


%% binding(Name, Bindings)
%% add_anon(Value, Bindings)

%% Try match bindings in the function head instead of
%% recursion on every Binding. Recursion after the 6:th element.

binding(Name, [{Name,Val}|_]) ->
    {value,Val};
binding(Name, [_,{Name,Val}|_]) ->
    {value,Val};
binding(Name, [_,_,{Name,Val}|_]) ->
    {value,Val};
binding(Name, [_,_,_,{Name,Val}|_]) ->
    {value,Val};
binding(Name, [_,_,_,_,{Name,Val}|_]) ->
    {value,Val};
binding(Name, [_,_,_,_,_,{Name,Val}|_]) ->
    {value,Val};
binding(Name, [_,_,_,_,_,_|Bs]) ->
    binding(Name, Bs);
binding(Name, [_,_,_,_,_|Bs]) ->
    binding(Name, Bs);
binding(Name, [_,_,_,_|Bs]) ->
    binding(Name, Bs);
binding(Name, [_,_,_|Bs]) ->
    binding(Name, Bs);
binding(Name, [_,_|Bs]) ->
    binding(Name, Bs);
binding(Name, [_|Bs]) ->
    binding(Name, Bs);
binding(_, []) ->
    unbound.

add_anon(Val, [{'_',_}|Bs]) ->
    [{'_',Val}|Bs];
add_anon(Val, [B1,{'_',_}|Bs]) ->
    [B1,{'_',Val}|Bs];
add_anon(Val, [B1,B2,{'_',_}|Bs]) ->
    [B1,B2,{'_',Val}|Bs];
add_anon(Val, [B1,B2,B3,{'_',_}|Bs]) ->
    [B1,B2,B3,{'_',Val}|Bs];
add_anon(Val, [B1,B2,B3,B4,{'_',_}|Bs]) ->
    [B1,B2,B3,B4,{'_',Val}|Bs];
add_anon(Val, [B1,B2,B3,B4,B5,{'_',_}|Bs]) ->
    [B1,B2,B3,B4,B5,{'_',Val}|Bs];
add_anon(Val, [B1,B2,B3,B4,B5,B6|Bs]) ->
    [B1,B2,B3,B4,B5,B6|add_anon(Val,Bs)];
add_anon(Val, [B1,B2,B3,B4,B5|Bs]) ->
    [B1,B2,B3,B4,B5|add_anon(Val,Bs)];
add_anon(Val, [B1,B2,B3,B4|Bs]) ->
    [B1,B2,B3,B4|add_anon(Val,Bs)];
add_anon(Val, [B1,B2,B3|Bs]) ->
    [B1,B2,B3|add_anon(Val,Bs)];
add_anon(Val, [B1,B2|Bs]) ->
    [B1,B2|add_anon(Val,Bs)];
add_anon(Val, [B1|Bs]) ->
    [B1|add_anon(Val,Bs)];
add_anon(Val, []) ->
    [{'_',Val}].

%% add_bindings(Bindings1, Bindings2)
%%  Add Bindings1 to Bindings2. Bindings in
%%  Bindings1 hides bindings in Bindings2.
%%  Used in list comprehensions (and funs).

add_bindings([{Name,V}|Bs], ToBs0) ->
    ToBs = add_binding(Name, V, ToBs0),
    add_bindings(Bs, ToBs);
add_bindings([], ToBs) ->
    ToBs.

add_binding(N, Val, [{N,_}|Bs]) ->
    [{N,Val}|Bs];
add_binding(N, Val, [B1,{N,_}|Bs]) ->
    [B1,{N,Val}|Bs];
add_binding(N, Val, [B1,B2,{N,_}|Bs]) ->
    [B1,B2,{N,Val}|Bs];
add_binding(N, Val, [B1,B2,B3,{N,_}|Bs]) ->
    [B1,B2,B3,{N,Val}|Bs];
add_binding(N, Val, [B1,B2,B3,B4,{N,_}|Bs]) ->
    [B1,B2,B3,B4,{N,Val}|Bs];
add_binding(N, Val, [B1,B2,B3,B4,B5,{N,_}|Bs]) ->
    [B1,B2,B3,B4,B5,{N,Val}|Bs];
add_binding(N, Val, [B1,B2,B3,B4,B5,B6|Bs]) ->
    [B1,B2,B3,B4,B5,B6|add_binding(N,Val,Bs)];
add_binding(N, Val, [B1,B2,B3,B4,B5|Bs]) ->
    [B1,B2,B3,B4,B5|add_binding(N,Val,Bs)];
add_binding(N, Val, [B1,B2,B3,B4|Bs]) ->
    [B1,B2,B3,B4|add_binding(N,Val,Bs)];
add_binding(N, Val, [B1,B2,B3|Bs]) ->
    [B1,B2,B3|add_binding(N,Val,Bs)];
add_binding(N, Val, [B1,B2|Bs]) ->
    [B1,B2|add_binding(N,Val,Bs)];
add_binding(N, Val, [B1|Bs]) ->
    [B1|add_binding(N,Val,Bs)];
add_binding(N, Val, []) ->
    [{N,Val}].

%% merge_bindings(Bindings1, Bindings2, Cm, Line, Function)
%%  Merge bindings detecting bad matches. 
%%  Special case '_', save the new one !!!
%%  Bindings1 is the newest bindings.

merge_bindings(Bs, Bs, _, _, _) ->
    Bs;                                         %Identical bindings
merge_bindings([{Name,V}|B1s], B2s, Cm, Line, F) ->
    case binding(Name, B2s) of
	{value,V} ->				%Already there, and the SAME
	    merge_bindings(B1s, B2s, Cm, Line, F);
	{value,V1} when Name == '_' ->		%	but anonym different
	    B2s1 = lists:keydelete('_',1,B2s),
	    [{Name,V}|merge_bindings(B1s, B2s1, Cm, Line, F)];
	{value,V1} ->				%	but different
	    exit({{badmatch,V},F},Cm,Line,B2s);
	unbound ->				%Not there, add it
	    [{Name,V}|merge_bindings(B1s, B2s, Cm, Line, F)]
    end;
merge_bindings([], B2s, _, _, _) ->
    B2s.

reverse([H1,H2,H3,H4,H5], S)   -> [H5,H4,H3,H2,H1|S];
reverse([H1,H2,H3,H4,H5|T], S) -> reverse(T, [H5,H4,H3,H2,H1|S]);
reverse([H1,H2,H3,H4], S)      -> [H4,H3,H2,H1|S];
reverse([H1,H2,H3], S)         -> [H3,H2,H1|S];
reverse([H1,H2], S)            -> [H2,H1|S];
reverse([H], S)                -> [H|S];
reverse([], S)                 -> S.

format_args(As) -> lists:flatten(format_args1(As)).

format_args1([A]) ->
    [io_lib:format('~p',[A])];
format_args1([A|As]) ->
    [io_lib:format('~p',[A]),$,|format_args1(As)];
format_args1([]) ->
    [].

trace_p() -> get(trace).

trace_receive(Le, TimeoutP) ->
    Tail = case TimeoutP of
	       true -> "with timeout ";
	       _    -> ""
	   end,
    case trace_p() of
	true -> trace_out(io_lib:format("++ (~w) receive " ++ Tail,[Le]));
	_    -> true
    end.

trace_received() ->
    case trace_p() of
	true -> trace_out(io_lib:format("~n",[]));
	_    -> true
    end.

trace_received(Msg) ->
    case trace_p() of
	true -> trace_out(io_lib:format("~n<== ~p~n",[Msg]));
	_    -> true
    end.

trace_fnk_call(Le,Line,M,Fu,As) ->
    case trace_p() of
	true -> trace_out(io_lib:format("++ (~w) <~w> ~w:~w(~s)~n",
					[Le,Line,M,Fu,format_args(As)]));
	_    -> true
    end.

trace_fnk_ret(Le,Val,false) ->
    case trace_p() of
	true -> trace_out(io_lib:format("-- (~w) ~p~n", [Le,Val]));
	_    -> true
    end;
trace_fnk_ret(_,_,_) ->
    true.

trace_bif_call(Le,Line,Name,As) ->
    case trace_p() of
	true -> trace_out(io_lib:format("++ (~w) <~w> ~w(~s)~n",
					[Le,Line,Name,format_args(As)]));
	_    -> true
    end.

trace_send(To,Msg) ->
    case trace_p() of
	true -> trace_out(io_lib:format("==> ~w : ~p~n",[To,Msg]));
	_    -> true
    end.

trace_out(String) ->
    dbg_icmd:tell_attached({self(),trace,String}).

safe_bif(register)           -> true;
safe_bif(unregister)         -> true;
safe_bif(whereis)            -> true;
safe_bif(registered)         -> true;
safe_bif(abs)                -> true;
safe_bif(float)              -> true;
safe_bif(trunc)              -> true;
safe_bif(round)              -> true;
safe_bif(math)               -> true;
safe_bif(node)               -> true;
safe_bif(length)             -> true;
safe_bif(hd)                 -> true;
safe_bif(tl)                 -> true;
safe_bif(size)               -> true;
safe_bif(element)            -> true;
safe_bif(setelement)         -> true;
safe_bif(atom_to_list)       -> true;
safe_bif(list_to_atom)       -> true;
safe_bif(integer_to_list)    -> true;
safe_bif(list_to_integer)    -> true;
safe_bif(float_to_list)      -> true;
safe_bif(list_to_float)      -> true;
safe_bif(tuple_to_list)      -> true;
safe_bif(list_to_tuple)      -> true;
safe_bif(make_ref)           -> true;
safe_bif(time)               -> true;
safe_bif(date)               -> true;
safe_bif(processes)          -> true;
safe_bif(process_info)       -> true;
safe_bif(load_module)        -> true;
safe_bif(delete_module)      -> true;
safe_bif(halt)               -> true;
safe_bif(check_process_code) -> true;
safe_bif(purge_module)       -> true;
safe_bif(pid_to_list)        -> true;
safe_bif(list_to_pid)        -> true;
safe_bif(module_loaded)      -> true;
safe_bif(binary_to_term)     -> true;
safe_bif(term_to_binary)     -> true;
safe_bif(alive)              -> true;
safe_bif(notalive)           -> true;
safe_bif(nodes)              -> true;
safe_bif(is_alive)           -> true;
safe_bif(disconnect_node)    -> true;
safe_bif(binary_to_list)     -> true;
safe_bif(list_to_binary)     -> true;
safe_bif(split_binary)       -> true;
safe_bif(concat_binary)      -> true;
safe_bif(term_to_atom)       -> true;
safe_bif(hash)               -> true;
safe_bif(pre_loaded)         -> true;
safe_bif(info)               -> true;
safe_bif(set_cookie)         -> true;
safe_bif(get_cookie)         -> true;
safe_bif(spawn)              -> spawn;
safe_bif(spawn_link)         -> spawn;
safe_bif(_)                  -> false.

