%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1992 Ellemtel Telecommunications Systems Laboratories
%% File    : dbg_iload.erl
%% Author  : Magnus Fr|berg
%% Purpose : Load a new version of a module to be
%%           interpreted.

-module(dbg_iload).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

%% ------------------------------------------------
%% -- Exported functions.
%% ------------------------------------------------

-export([load_mod/3,load_mod1/3]).

%%% =================================================
%%% Load a new module into the database.
%%%
%%% We want the loading of a module to be syncronous
%%% so no other process tries to interpret code in
%%% a module not being completely loaded. This is
%%% achieved as this function is called from the
%%% 'interpret' server.
%%% We are suspended until the module has been loaded.
%%% =================================================

load_mod(Mod,File,Binary) ->
    Flag = process_flag(trap_exit,true),
    Pid = spawn_link(?MODULE,load_mod1,[Mod,File,Binary]),
    receive
	{'EXIT',Pid,What} ->
	    process_flag(trap_exit,Flag),
	    What
    end.

load_mod1(Mod,File,Binary) ->
    store_module(Mod,File,Binary),
    int:int_module(Mod),
    exit({ok,Mod}).

store_module(Mod,File,Binary) ->
    {AbsBin,FileBin} = binary_to_term(Binary),
    AbstractList = binary_to_term(AbsBin),
    FileList = binary_to_list(FileBin),
    store_mod1(Mod,File,FileList,AbstractList).

store_mod1(Mod,File,FileList,AbstractList) ->
    Db = dbg_idb:cr_new_module(Mod),
    dbg_idb:insert(Db,mod_file,File),
    store_abs(Mod,Db,AbstractList,false),
    NewBinary = store_mod_line_no(Mod,Db,FileList),
    dbg_idb:insert(Db,mod_bin,NewBinary),
    dbg_idb:insert(Db,module,Mod).

store_abs(_,_,[],_) -> true;
store_abs(Mod,Db,[Form|Forms],Export_all) ->
    ExpAll = store_mod(Mod,Db,Form,Export_all),
    store_abs(Mod,Db,Forms,ExpAll).

store_mod(Mod,Db,{exports,Value},Exp) ->
    dbg_idb:insert(Db,export,Value),
    Exp;
store_mod(Mod,Db,{attribute,_,export,Value},Exp) ->
    case dbg_idb:lookup(Db,export) of
	{ok,OldVal} ->
	    dbg_idb:insert(Db,export,
			  lists:append(Value,OldVal));
	_ ->
	    dbg_idb:insert(Db,export,Value)
    end,
    Exp;
store_mod(Mod,Db,{attribute,_,compile,export_all},_) ->
    true;
store_mod(Mod,Db,{attribute,_,compile,Value},Exp) ->
    case lists:member(export_all,Value) of
	true -> true;
	_    -> Exp
    end;
store_mod(Mod,Db,{attribute,_,_,_},Exp) ->
    Exp;
store_mod(Mod,Db,{function,_,Fnk,Arity,Cs},false) ->
    dbg_idb:insert(Db,{Fnk,Arity},Cs),
    false;
store_mod(Mod,Db,{function,_,Fnk,Arity,Cs},true) ->
    dbg_idb:insert(Db,{Fnk,Arity},Cs),
    case dbg_idb:lookup(Db,export) of
	{ok,Exported} ->
	    case lists:member({Fnk,Arity},Exported) of
		false ->
		    dbg_idb:insert(Db,export,
				  [{Fnk,Arity}|Exported]);
		_ ->
		    true
	    end;
	_ ->
	    dbg_idb:insert(Db,export,[{Fnk,Arity}])
    end,
    true;
store_mod(Mod,Db,{defs,Defs},Exp) -> %% Internal repr. for macro definitions
    dbg_idb:insert(Db,defs,Defs),
    Exp;
store_mod(Mod,_,Form,_) ->
    io:format('Unrecognized form during parse~n',[]),
    exit({unknown_form,Form}).

make_import_list(From,[{Func,Arity}|Funcs]) ->
    [{From,{Func,Arity}}|make_import_list(From,Funcs)];
make_import_list(_,[]) ->
    [].

store_mod_line_no(Mod,Db,Contents) ->
    store_mod_line_no(Mod,Db,Contents,1,0,[]).

store_mod_line_no(Mod,_,[],LineNo,Pos,NewCont) ->
    list_to_binary(lists:append(lists:reverse(NewCont)));
store_mod_line_no(Mod,Db,Contents,LineNo,Pos,NewCont) when integer(LineNo) ->
    {ContTail,Pos1,NewCont1} = store_line(Mod,Db,Contents,LineNo,Pos,NewCont),
    store_mod_line_no(Mod,Db,ContTail,LineNo+1,Pos1,NewCont1).

store_line(Mod,Db,Contents,LineNo,Pos,NewCont) ->
    S = string:left(lists:flatten(io_lib:format("~w:",[LineNo])),8,$ ),
    {ContHead,ContTail,PosNL} = get_nl(Contents,Pos+8,[]),
    dbg_idb:insert(Db,LineNo,{Pos+8,PosNL}),
    {ContTail,PosNL+1,[lists:append(S,ContHead)|NewCont]}.

get_nl([10|T],Pos,Head) -> {lists:reverse([10|Head]),T,Pos};
get_nl([H|T],Pos,Head) ->
    get_nl(T,Pos+1,[H|Head]);
get_nl([],Pos,Head) -> {lists:reverse(Head),[],Pos}.
    

