%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%----------------------------------------------------------------------
%%% File    : pman_buf.erl
%%% Author  : Peter Olin <olin@erlang.ericsson.se>
%%% Purpose : This module is the exported interface to the buffering mechanism
%%%           used in PMAN to prevent the trace output to congest
%%%           the system.
%%%
%%%           This module depends on the modules (direct or indirectly):
%%%               pman_buf.hrl
%%%               pman_buf_utils.erl
%%%               pman_buf_buffer.erl
%%%               pman_buf_converter.erl
%%%               pman_buf_printer.erl
%%%
%%% Created :  3 Jun 1997 by Peter Olin <olin@erlang.ericsson.se>
%%%----------------------------------------------------------------------

-module(pman_buf).
-author('olin@erlang.ericsson.se').

%%-compile(export_all).
-export([start/2,clear/3]).


-include("pman_buf.hrl").


%% The buffering mechanism consists of three processes that 
%% work in a chain to prevent the process monitor from congesting
%% output mechanism.
%% 
%% Messages are buffered in the CONVERTER process before they are sent to
%% to the BUFFER process where they are formatted before they are finally
%% sent to either a file or the PRINTER process. The printer process
%% outputs the messages in the graphical user interface.
%%
%%
%%
%%     	   -->   CONVERTER  -->   BUFFER    --> PRINTER --> gui
%%                                 |
%%                                 |
%%                                 |
%%                                 V
%%
%%                                file
%%





%% ----------------------------------------------------------------
%% The amount of data produced by a trace message may be large, and
%% cause the run time system to run out of memory. To avoid this,
%% the task of storing, cutting buffers, formating data and printing
%% it is performed by three processes: The buffer, the converter and 
%% the printer.
%%
%% The converter accepts the raw data, a list
%% of {trace,Msg} tuples. Having max priority, it assures that the
%% amount of raw data stored never excedes ?BUFF_SIZE messages.
%% (With the exception of the last batch received, which assures that
%% the last trace message printed is never a buffer cut message.)
%% Whenever there is space available in the buffer process, (The
%% Buffer process stores max. ?BUFF_SIZE converted messages),
%% the buffer asks for more unconverted messages, and ?PRINT_LEN messages 
%% are sent. They are converted by the buffer, and added to the list
%% of messages to be sent. 

%% The printer process requests formatted messages from the buffer,
%% and in chuncs of ?MAX_OUTPUT sends them to the buffer. If traces
%% are to be dumped on file, due to the max priority, such is handled
%% in the converter, and buffers are not cut.
%%


%% ---------------------------------------------------------------
%% Initializes the buffering mechanism, which consist of three
%% processes, each involved with a phase of the formattation and
%% output of data to the process windows.

start(Editor, FileName) ->
    Buffer_Pid = spawn_link(pman_buf_buffer,init,[Editor]),
    Converter_Pid =
	spawn_link(pman_buf_converter,init,[Buffer_Pid, FileName]),
    Buffer_Pid!{converter_pid, Converter_Pid},
    #buffer{converter=Converter_Pid,buffer=Buffer_Pid}.



%% ---------------------------------------------------------------
%% Kills the converter and the clears the buffer with formated data
%% starting a new converter.

clear(Buff,String, FileName) ->
    exit(Buff#buffer.converter,win_killed),
    Converter_Pid=spawn_link(pman_buf_converter,init,[Buff#buffer.buffer,
						      FileName]),
    Buff#buffer.buffer!{clear,String,Converter_Pid },
    Buff#buffer{converter = Converter_Pid}.



