%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%----------------------------------------------------------------------
%%% File    : pman_relay.erl
%%% Author  : Peter Olin <olin@erlang.ericsson.se>
%%% Purpose : Interface function to  relay calls (esp. trace calls) 
%%%           to processes on other  nodes. Some of the calls
%%%           are conditionally relayed.
%%% Created : 31 Oct 1997 by Peter Olin <olin@erlang.ericsson.se>
%%%----------------------------------------------------------------------

-module(pman_relay).
-author('olin@erlang.ericsson.se').

%%-compile(export_all).
-export([start/1,
	 ok_to_trace/1,
	 trac/3]).


-include("assert.hrl").

%% --------------------------------------------------------------
%%                  DISTRIBUTION
%% --------------------------------------------------------------
%% (???) Process dictionary alert!!!
%%
%% Since we are not allowed to do erlang:trace/3 on remote
%% processe we create a help process at the remote node to
%% do the job for us 
%% ---------------------------------------------------------------

start(P) when pid(P), node(P) /= node() ->
    
    %% Remote supervision, relaying necessary
    
    put(relay, spawn_link(node(P), pman_relay_server, init, [self()]));


start(_) ->
    
    %% Local supervision, no relaying
    
    ignore.




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ok_to_trace/1 - Tests wheter we can actually start tracing 
%%    a process.
%%
%% Arguments:
%%   Pid		Pid of the process to trace (on local or remote node)
%%
%% Returns
%%   true		If it is OK to trace the process
%%   false		If the process is already traced, or some other
%%			condition prevents it from being traced.

ok_to_trace(Pid) when node(Pid) == node()->

    %% Local trace, no relaying

    case catch erlang:trace(Pid, false, [send]) of
	true ->
	    true;
	_Otherwise ->
	    false
    end;
ok_to_trace(Pid) ->

    %% Remote trace, relaying necessary

    PidRelay =  get(relay),
    PidRelay ! {ok_to_trace, self(), Pid},
    receive
	{ok_to_trace, PidRelay} ->
	    true;
	{not_ok_to_trace, PidRelay} ->
	    false;
	_Otherwise ->
	    ?ALWAYS_ASSERT("Unexpected message from relay process")
    after
	5000 ->
	    false
    end.


    
	     
    

%% ---------------------------------------------------------------
%% Possibly send a request to do tracing to a remote node.
%% ---------------------------------------------------------------

trac(Pid, How, Flag) when node(Pid) == node() ->

    %% Local trace, no relaying necessary


    case catch erlang:trace(Pid, How, Flag) of
	true -> ok;
	_ -> pman_win:format("** Illegal trace request ** \n", [])
    end;
	
trac(Pid, How, Flag) ->



    %% Remote trace, relaying necessary

    get(relay) ! {self(), erlang, trace, [Pid, How, Flag]}.

