%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(toolbar_lib).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('Gunilla Hugosson <gunilla@erlang.ericsson.se>').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Erlang Toolbar
%
%%% Description %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Library functions
%
%%% Exports %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
-export([gs_contribs_dir/0,help_file/0,otp_file/0]).
-export([error_string/1]).
-export([legal_file/1]).
-export([insert_newlines/1]).
-export([tool_info_syntax/2]).
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Exported functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%----------------------------------------
% gs_contribs_dir() => string()
% Return the directory of the GS contributions
%----------------------------------------
gs_contribs_dir() ->
    filename:join(gtk_port_handler:gs_dir(),"../contribs/ebin/").

%----------------------------------------
% help_file() => string()
% Returns the address to the toolbar help file
%----------------------------------------
help_file() ->
    filename:join(filename:dirname(code:which(toolbar)),
		 "../doc/index.html").

%----------------------------------------
% otp_file() => string()
% Returns the address to the OTP documentation
%----------------------------------------
otp_file() ->
    filename:join(code:root_dir(),"doc/index.html").

%----------------------------------------
% error_string(Reason) => string()
%   Reason - nofile | format | noname | nostart | illegal | raccess | waccess
% Given Reason, returns a short "explanation string"
%----------------------------------------
error_string(nofile) -> "File does not exist";
error_string(format) -> "File on wrong format";

error_string(noname) -> "No tool name is specified";
error_string(nostart) -> "No start function is specified";

error_string(illegal) -> "Illegal file name";

error_string(raccess) ->  "File is not readable";
error_string(waccess) ->  "File is not writeable".

%----------------------------------------
% legal_file(File) => ok | directory | {error,nofile} | {error,raccess}
%   File - string() File name
% Checks if File is an existing and readable file.
%----------------------------------------
legal_file(File) ->
    case file:file_info(File) of

	%% File exists...
	{ok,{_Size,regular,Access,_LastAccess,_LastModify,_,_}} ->
	    if
		
		%% ...but is read protected
		Access/=read,
		Access/=read_write ->
		    {error,raccess};
		
		%% ...and is possible to read
		true ->
		    ok
	    end;
	
	%% File does not exist
	Error -> % {error,file_info}
	    case file:list_dir(File) of
		{ok,_FileNameList} ->
		    directory;
		Error ->
		    {error,nofile}
	    end
    end.

%----------------------------------------
% insert_newlines(Strings) => string()
%   Strings - string() | [string()]
% If Strings is a list of strings, return a string where all these strings
% are concatenated with newlines in between, otherwise return Strings.
%----------------------------------------
insert_newlines([String|Rest]) when list(String), Rest/=[]->
    String ++ "\n" ++ insert_newlines(Rest);
insert_newlines([Last]) ->
    [Last];
insert_newlines(Other) ->
    Other.

%----------------------------------------
% tool_info_syntax(Version,ToolInfo) => {ok,CorrToolInfo} | {error,Reason}
%   Version - string()
%   ToolInfo - tuple()
%   CorrToolInfo - list()
%   Reason - version | format | noname | nostart
% Return a corrected (blanks removed etc) version of ToolInfo
% if the syntax of ToolInfo is correct (ie contains all
% mandatory elements and all values are of the specified type).
%
% Currently accepted versions:
%   "0.1" (which should be "1.0")
%   "1.1" (same as "0.1")
%----------------------------------------
tool_info_syntax("1.1",ToolInfo) ->
    tool_info_syntax("0.1",ToolInfo);
tool_info_syntax("0.1",ToolInfo) when tuple(ToolInfo) ->
    syntax01(tuple_to_list(ToolInfo),false,false,[]);
tool_info_syntax("0.1",_) ->
    {error,format};
tool_info_syntax(Vsn,_) ->
    {error,version}.

%%% Internal functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%----------------------------------------
% syntax01(List,NameF,StartF,Res) => {ok,Res} | {error,Reason}
%   List - [tuple()]
%   NameF, StartF - boolean()
%   Res - [tuple()]
%   Reason - format | noname | nostart
% Version 0.1 syntax check of .tool file.
%----------------------------------------
syntax01([{tool,Str}|Rest],false,StartF,Res) when list(Str) ->
    case string:strip(Str) of
	[] ->
	    {error,format};
	Tool ->
	    syntax01(Rest,true,StartF,[{tool,Tool}|Res])
    end;
syntax01([{start,{M,F,A}}|Rest],NameF,false,Res) when atom(M),
						      atom(F),
						      list(A) ->
    syntax01(Rest,NameF,true,[{start,{M,F,A}}|Res]);
syntax01([{icon,Str}|Rest],NameF,StartF,Res) when list(Str) ->
    case string:strip(Str) of
	[] ->
	    syntax01(Rest,NameF,StartF,Res);
	Icon ->
	    syntax01(Rest,NameF,StartF,[{icon,Icon}|Res])
    end;
syntax01([{message,Str}|Rest],NameF,StartF,Res) when list(Str) ->
    case string:strip(Str) of
	[] ->
	    syntax01(Rest,NameF,StartF,Res);
	Message ->
	    syntax01(Rest,NameF,StartF,
		     [{message,lists:sublist(Message,1,30)}|Res])
    end;
syntax01([{html,Str}|Rest],NameF,StartF,Res) when list(Str) ->
    case string:strip(Str) of
	[] ->
	    syntax01(Rest,NameF,StartF,Res);
	Html ->
	    syntax01(Rest,NameF,StartF,[{html,Html}|Res])
    end;
syntax01([],true,true,Res) ->
    {ok,Res};
syntax01([],false,_,_) ->
    {error,noname};
syntax01([],_,false,_) ->
    {error,nostart};
syntax01(_,_,_,_) ->
    {error,format}.
