%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%*********************************************************************
%%%
%%%   Filename:         tv_pc.erl
%%%
%%%   Author:           EPK/TE  Fredrik Gustafson 
%%% 
%%%   SCCS Delta ID:    1.3
%%%
%%%   SCCS Delta Date:  97/06/12
%%%
%%%   SCCS ID:          @(#)  tv_pc.erl  1.3  97/06/12
%%%
%%%   Description:      pc part of the table tool, i.e., the process controlling
%%%                     all other processes, and managing the actions to take.
%%%
%%%*********************************************************************
%%%
%%%   COPYRIGHT ERICSSON TELECOM AB 1998
%%%
%%%   THE COPYRIGHT TO THE COMPUTER PROGRAM HEREIN IS THE
%%%   PROPERTY OF ERICSSON TELECOM AB, SWEDEN. THE PROGRAM
%%%   MAY BE USED AND/OR COPIED ONLY WITH THE WRITTEN PERMISSION
%%%   FROM ERICSSON TELECOM AB, OR IN ACCORDANCE WITH THE TERMS
%%%   AND CONDITIONS STIPULATED IN THE AGREEMENT/CONTRACT UNDER
%%%   WHICH THE PROGRAM HAS BEEN SUPPLIED.
%%%
%%%*********************************************************************
%%%
%%%   Revision information:
%%%
%%%   Date      Name                       Description
%%%   ------    -----------------------    -----------
%%%   970427    EPK/TE  F Gustafson        First version.
%%%
%%%
%%%*********************************************************************



-module(tv_pc).
-author('fredrik@erlang.ericsson.se').
-vsn('$Revision: /main/release/free/1').
-copyright('Copyright (C) 1991-98 Ericsson Telecom AB').



-export([pc/0,
	send_data/2
	]).



-include("tv_int_def.hrl").
-include("tv_int_msg.hrl").
-include("tv_pc_int_def.hrl").




%%%*********************************************************************
%%% EXTERNAL FUNCTIONS
%%%*********************************************************************



%%======================================================================
%% Function:      pc.
%%
%% Return Value:  None.
%%
%% Description:   Process controlling the processes 'pd', 'pw', 'dbs' and 'etsread'.
%%                After necessary initialisations, an eternal loop is
%%                entered, where window created messages are received and 
%%                handled, as well as user input.
%%
%% Parameters:    None.
%%======================================================================



pc() ->
    process_flag(trap_exit, true),
    ProcVars = start_procs(#process_variables{}),
    loop(ProcVars).







%%%********************************************************************
%%% INTERNAL FUNCTIONS
%%%********************************************************************





%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


start_procs(ProcVars) ->
    IpPid = spawn(tv_ip, ip, [self()]),
    
    show_progress(IpPid, 4, "Initializing database..."),

    PwPid = spawn_link(tv_pw, pw, [self()]),
    sleep(1500),

    show_progress(IpPid, 2, "Loading..."),
    sleep(500),

    PdPid = spawn_link(tv_pd, pd, [self()]),

    show_progress(IpPid, 2, "Loading..."),
    sleep(500),

    DbsPid = spawn_link(tv_db, dbs, [self()]),

    show_progress(IpPid, 2, "Loading..."),
    sleep(500),
    show_progress(IpPid, 2, "Starting table interface..."),
    sleep(500),
    show_progress(IpPid, 2, "Starting table interface..."),

    EtsreadPid = spawn_link(tv_etsread, etsread, [self()]),

    sleep(500),
    show_progress(IpPid, 2, "Starting table interface..."),

    sleep(500),

    NewWinP = init_pw(PwPid, ProcVars),

    show_progress(IpPid, 4, "Initializing graphics..."),

    init_pd(PdPid, NewWinP),

    show_progress(IpPid, 4, "Initializing graphics..."),
    sleep(500),
    show_progress(IpPid, 5, "Loading..."),
    sleep(250),

    IpPid ! #ip_quit{sender = self()},
    sleep(500),

       % Now make window visible!
    gs:config(NewWinP#window_params.window_id, [{map, true}]),

    ProcVars#process_variables{pw_pid        = PwPid,
			       pd_pid        = PdPid,
			       dbs_pid       = DbsPid,
			       etsread_pid   = EtsreadPid,
			       current_node  = node(),
			       window_params = NewWinP
			       }.







%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


show_progress(IpPid, NofElements, Text) ->
    IpPid ! #ip_update{sender               = self(),
		       nof_elements_to_mark = NofElements,
		       text                 = Text
		      }.

    





%%======================================================================
%% Function:      loop.
%%
%% Return Value:  None.
%%
%% Description:   Eternal (well, almost) loop, receiving messages and 
%%                handling them.
%%
%% Parameters:    None.
%%======================================================================


loop(ProcVars) ->
    receive
	Msg ->
	    case Msg of 

                   % Normal messages!
		#dbs_subset{} ->
		    send_data(Msg, ProcVars),
		    NewProcVars = check_time_to_poll_table(Msg, ProcVars),
		    loop(NewProcVars);

		#pc_poll_table{} ->
		    TempProcVars = ?MENU_FUNC_FILE:poll_table(ProcVars),
		    NewProcVars  = check_node(TempProcVars),
		    loop(NewProcVars);

		#pc_open_table{} ->
		    NewProcVars = ?MENU_FUNC_FILE:open_table(ProcVars),
		    loop(NewProcVars);

		#pc_search_req{} ->
		    DbsPid = ProcVars#process_variables.dbs_pid,
		    DbsPid ! #dbs_search_req{sender=self()},
		    loop(ProcVars);

		#pc_set_sorting_mode{} ->
		    set_sorting_mode(Msg, ProcVars),
		    loop(ProcVars);


                #pc_data_req{element = Pos, nof_elements = Length} ->
		    DbsPid = ProcVars#process_variables.dbs_pid,
		    DbsPid ! #dbs_subset_req{sender           = self(), 
					     subset_pos       = Pos,
					     subset_length    = Length
					    },
                    loop(ProcVars);


		#pc_marked_row{row_no = RowNo} ->
		    DbsPid = ProcVars#process_variables.dbs_pid,
		    DbsPid ! #dbs_marked_row{sender = self(),
					     row_no = RowNo
					    },
		    loop(ProcVars);
		

                #pc_menu_msg{} ->
                    Fcn = Msg#pc_menu_msg.data,
                    NewProcVars = {?MENU_FUNC_FILE, Fcn}(ProcVars),
                    loop(NewProcVars);
 
%		#pc_resend_data{} ->
%                    NewProcVars = resend_data(ProcVars),
%                    loop(NewProcVars);
		
		#pc_show_table_info{} ->
		    NewProcVars = ?MENU_FUNC_FILE:table_info(ProcVars),
		    loop(NewProcVars);

                #pc_win_conf{} ->
                    NewProcVars = ?GRAPH_FUNC_FILE:win_conf(Msg, ProcVars),
                    loop(NewProcVars);
		
		#pc_help{} ->
		    NewProcVars = ?MENU_FUNC_FILE:help_button(ProcVars),
		    loop(NewProcVars);

		#pc_dead_table{automatic_polling = AutoPoll} ->
		    WinP  = ProcVars#process_variables.window_params,
		    WinId = WinP#window_params.window_id,
		    gs:config(WinId, [beep]),
		    tool_utils:notify(WinId, ["The table no longer exists!"]),
		    NewProcVars =
			case AutoPoll of
			    true ->
				tool_utils:notify(WinId,
						  ["The automatic polling is turned off!"]),
				ProcVars#process_variables{poll_interval = infinity};
			    false ->
				ProcVars
			end,
		    loop(NewProcVars);

		#pc_nodedown{automatic_polling = AutoPoll} ->
		    WinP = ProcVars#process_variables.window_params,
		    WinId = WinP#window_params.window_id,
		    gs:config(WinId, [beep]),
		    tool_utils:notify(WinId, ["The remote node is down!"]),
		    NewProcVars = 
			case AutoPoll of
			    true ->
				tool_utils:notify(WinId,
						  ["The automatic polling is turned off!"]),
				ProcVars#process_variables{poll_interval = infinity};
			    false ->
				ProcVars
			end,
		    loop(NewProcVars);
		    

		   % Exit messages!
		{'EXIT', Sender, Reason} ->
		    exit_signals({Sender, Reason}, ProcVars);
		

		_Other ->
		    loop(ProcVars)

	    end
    end.
	





check_node(ProcVars) ->
    #process_variables{pw_pid       = PwPid,
		       current_node = OldCurrNode,
		       table_id     = TableId,
		       table_type   = TableType,
		       table_name   = TableName,
		       table_number = TableNumber}  = ProcVars,

    HomeNode = node(),
    case net_adm:ping(OldCurrNode) of
	pong ->
	    ProcVars;
	pang when OldCurrNode /= 'nonode@nohost', HomeNode /= 'nonode@nohost' ->
	    ProcVars;
	pang ->
	       %% Two possibilities:
	       %%   1. the old current node was 'nonode@nohost'. If it no longer exists,
	       %%      then the system has become distributed.
	       %%   2. the old current node no longer exists. If the new local node
	       %%      is 'nonode@nohost', the system no longer is distributed.
	    WinTitle = ?MENU_FUNC_FILE:get_window_title(HomeNode, TableId, TableType,
							TableName, TableNumber),

	    PwPid ! #pw_set_window_title{sender = self(),
					 win_title = WinTitle},

	    ProcVars#process_variables{current_node = HomeNode}
    end.

	    
	    

    








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


send_data(Msg, ProcVars) ->
    PdPid = ProcVars#process_variables.pd_pid,
    #dbs_subset{data                  = DbData,
		subset_pos            = ScalePos,
		db_length             = DbLength,
		list_of_keys          = ListOfKeys,
		max_elem_size         = MaxElemSize, 
		requested_row         = ReqRowData}  = Msg,

    Range = case ScalePos of
		0 ->
		    {0, 0};
		_Other ->
		    {1, DbLength}
	    end,

    PdPid ! #pc_data{sender        = self(),
		     scale_pos     = ScalePos, 
		     scale_range   = Range, 
		     elementlist   = DbData,
		     list_of_keys  = ListOfKeys,
		     max_elem_size = MaxElemSize,
		     marked_row    = ReqRowData
		    }.








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


set_sorting_mode(Msg, ProcVars) ->
    #pc_set_sorting_mode{sorting     = Sorting,
			 reverse     = Reverse,
			 sort_key_no = SortKeyNo} = Msg,

    DbsPid    = ProcVars#process_variables.dbs_pid,
    PdPid     = ProcVars#process_variables.pd_pid,
    TableType = ProcVars#process_variables.table_type,
    NewSortKeyNo = case SortKeyNo of 
		       undefined ->
			   if
			       TableType == mnesia ->
				   2;
			       true ->
				   1
			   end;
		       _Other ->
			   SortKeyNo
		   end,
    
    DbsPid ! #dbs_sorting_mode{sender      = self(),
			       sorting     = Sorting,
			       reverse     = Reverse,
			       sort_key_no = NewSortKeyNo
			      },

    PdPid ! #pc_set_sorting_mode_cfm{sender = self(),
				     sort_key_no = NewSortKeyNo
				    }.







%%======================================================================
%% Function:      init_pw.
%%
%% Return Value:  Tuple containing the Pid of the pw process, and the id of 
%%                the window created by the pw process.
%%
%% Description:   Starts the pw process, and orders it to create a window.
%%                (The size of the window may be given as option.)
%%
%% Parameters:    None.
%%======================================================================



init_pw(PwPid, ProcVars) ->
    #process_variables{window_params = WinP}  = ProcVars,

       % Now deblock pw, and order it to create a window!
    PwPid ! #pw_deblock{sender         = self(),
			win_title      = ?APPLICATION_NAME,
			win_width      = ?DEFAULT_WINDOW_WIDTH,
			win_height     = ?DEFAULT_WINDOW_HEIGHT,
			min_win_width  = ?DEFAULT_MIN_WINDOW_WIDTH,
			min_win_height = ?DEFAULT_MIN_WINDOW_HEIGHT
		       },
    

    receive
	#pw_deblock_cfm{win_id = WindowId} ->
	    ?MENU_FUNC_FILE:create_menus(PwPid),
	    
	       % Store the window id as well as the size of it.
	    WinP#window_params{window_id     = WindowId,
			       window_width  = ?DEFAULT_WINDOW_WIDTH,
			       window_height = ?DEFAULT_WINDOW_HEIGHT
			      }

    
    after 180000 ->      % A timeout of 1000 ms is too short, at least the first
	                % time the system is started!
	    exit(error)
    end.
	    
		    








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


init_pd(PdPid, WinP) ->
    #window_params{window_id     = WindowId,
		   window_width  = WindowWidth,
		   window_height = WindowHeight} = WinP,
    
       % Now deblock pd, and order it to create a canvas and a scale!
    PdPid ! #pd_deblock{sender      = self(),
			 win        = WindowId,
			 win_width  = WindowWidth,
			 win_height = WindowHeight,
			 scale      = true
			},

    receive
	#pd_deblock_cfm{} ->
	    done
    after 180000 ->
	    exit(error)
    end.
	    
		    


    





%%======================================================================
%% Function:      exit_signals.
%%
%% Return Value:  None.
%%
%% Description:   Decides, given an error message, action to take, i.e., whether
%%                operation shall procede, any process shall be restarted, or
%%                the table tool terminated.
%%
%% Parameters:    Exit_info:  tuple containing sender of the error message, and the 
%%                reason.
%%======================================================================


exit_signals(Exit_info, ProcVars) ->
    #process_variables{pd_pid      = PdPid,
		       pw_pid      = PwPid,
		       dbs_pid     = DbsPid,
		       etsread_pid = EtsreadPid} = ProcVars,

    case Exit_info of
        {PwPid, normal} ->
            exit(normal);

	{PwPid, error} ->
	    io:format("Internal error... restarting. ~n"),
	    kill_procs(normal, [PdPid, EtsreadPid, DbsPid]),
	    NewProcVars = start_procs(ProcVars),
	    loop(NewProcVars);	    

	{PdPid, _Reason} ->
	    io:format("Internal error... restarting. ~n"),
	    kill_procs(normal, [PwPid, EtsreadPid, DbsPid]),
	    NewProcVars = start_procs(ProcVars),
	    loop(NewProcVars);
	
	{DbsPid, _Reason} ->
	    io:format("Internal error... restarting. ~n"),
	    kill_procs(normal, [PdPid, PwPid, EtsreadPid]),
	    NewProcVars = start_procs(ProcVars),
	    loop(NewProcVars);
	
	{EtsreadPid, _Reason} ->
	    io:format("Internal error... restarting. ~n"),
	    kill_procs(normal, [PdPid, PwPid, DbsPid]),
	    NewProcVars = start_procs(ProcVars),
	    loop(NewProcVars);
	
        {_Sender, _OtherReason} ->
	    loop(ProcVars)
    end.









%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


kill_procs(_Status, []) ->
    done;
kill_procs(Status, [Pid | Tail]) ->
    exit(Pid, Status),
    kill_procs(Status, Tail).








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


sleep(T) ->
    receive 
    after T ->
	    done
    end.
    






%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


check_time_to_poll_table(Msg, ProcVars) ->
    #dbs_subset{required_time_etsread = EtsreadTime,
		required_time_dbs     = DbsTime}  = Msg,

    UserSetPollInterval = ProcVars#process_variables.poll_interval,
    WinP                = ProcVars#process_variables.window_params,
    WinId               = WinP#window_params.window_id,
    
    case too_short_pollinterval_chosen(UserSetPollInterval, EtsreadTime, DbsTime) of
	true ->
	    EtsreadPid   = ProcVars#process_variables.etsread_pid,
	    EtsreadPid ! #etsread_set_poll_interval{sender   = self(),
						    interval = infinity},
	    
	    TimeRequired = trunc(max_time_required(EtsreadTime, DbsTime) / 10 + 0.5) * 10 + 20,
	    
	    gs:config(WinId, [beep]),
	    tool_utils:notify(WinId, ["The current poll interval is too short!"]),
	    Str = "to " ++ lists:flatten(io_lib:write(TimeRequired)) ++ " seconds!",
	    tool_utils:notify(WinId, ["Setting the poll interval", Str]),
	    clear_message_buffer(),
	    EtsreadPid ! #etsread_set_poll_interval{sender   = self(),
						    interval = TimeRequired},
	    
	    ProcVars#process_variables{poll_interval = TimeRequired};
	false ->
	    ProcVars
    end.






clear_message_buffer() ->
    receive 
	#dbs_subset{} ->
	    clear_message_buffer()
    after 100 ->
	    done
    end.





max_time_required(T1, T2) when number(T1), number(T2) ->
    if
	T1 > T2 ->
	    T1;
	true ->
	    T2
    end;
max_time_required(T1, T2) when number(T1) ->
    T1;
max_time_required(T1, T2) ->
    T2.
    







%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


too_short_pollinterval_chosen(infinity, _EtsreadTime, _DbsTime) ->
    false;
too_short_pollinterval_chosen(undefined, _EtsreadTime, _DbsTime) ->
    false;
too_short_pollinterval_chosen(PollInt, EtsreadTime, _DbsTime) when EtsreadTime >= PollInt, number(EtsreadTime) ->
    true;
too_short_pollinterval_chosen(PollInt, EtsreadTime, DbsTime) when DbsTime >= PollInt, number(DbsTime) ->
    true;
too_short_pollinterval_chosen(PollInt, EtsreadTime, DbsTime) ->
    false.

