%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%*********************************************************************
%%%
%%%   Filename:         tv_pd_display.erl
%%%
%%%   Author:           EPK/TE  Fredrik Gustafson 
%%% 
%%%   SCCS Delta ID:    1.2
%%%
%%%   SCCS Delta Date:  97/04/28
%%%
%%%   SCCS ID:          @(#)  tv_pd_display.erl  1.2  97/04/28
%%%
%%%   Description:      Part of pd controlling the graphics.
%%%
%%%*********************************************************************
%%%
%%%   COPYRIGHT ERICSSON TELECOM AB 1998
%%%
%%%   THE COPYRIGHT TO THE COMPUTER PROGRAM HEREIN IS THE
%%%   PROPERTY OF ERICSSON TELECOM AB, SWEDEN. THE PROGRAM
%%%   MAY BE USED AND/OR COPIED ONLY WITH THE WRITTEN PERMISSION
%%%   FROM ERICSSON TELECOM AB, OR IN ACCORDANCE WITH THE TERMS
%%%   AND CONDITIONS STIPULATED IN THE AGREEMENT/CONTRACT UNDER
%%%   WHICH THE PROGRAM HAS BEEN SUPPLIED.
%%%
%%%*********************************************************************
%%%
%%%   Revision information:
%%%
%%%   Date      Name                       Description
%%%   ------    -----------------------    -----------
%%%   970427    EPK/TE  F Gustafson        First version.
%%%
%%%
%%%*********************************************************************



-module(tv_pd_display).
-author('fredrik@erlang.ericsson.se').
-vsn('$Revision: /main/release/free/1').
-copyright('Copyright (C) 1991-98 Ericsson Telecom AB').




-export([init_display/4, 
	 display_data/8, 
	 resize_display/3, 
	 resize_column/4, 
	 scroll_horizontally/2, 
	 scroll_vertically/2, 
	 perform_horizontal_scroll/2,
	 perform_vertical_scroll/2, 
	 marked_cell/5, 
	 update_toolbar_label/4,
	 get_data_element/4, 
	 hide_toolbar_editor/1, 
	 show_toolbar_editor/1]).





-include("tv_int_def.hrl").
-include("tv_int_msg.hrl").
-include("tv_pd_int_def.hrl").
-include("tv_pd_int_msg.hrl").







%%%*********************************************************************
%%% EXTERNAL FUNCTIONS
%%%*********************************************************************




%%======================================================================
%% Function:      init_display.
%%
%% Return Value:  Id of the display (here:canvas) created.
%%
%% Description:   Creates the canvas and the scale.
%%
%% Parameters:    Id of the window the display shall be created in.
%%======================================================================


init_display(WindowId, WindowWidth, WindowHeight, ProcVars) ->
       % Get all necessary window parameters!
    #process_variables{pg_pid         = PgPid,
		       pb_pid         = PbPid,
		       frame_params   = FrameP,
		       scale_params   = ScaleP,
		       toolbar_params = ToolP}  = ProcVars,
    
    NewFrameP = tv_pd_frames:create_display_frames(WindowId, WindowWidth,
						    WindowHeight, FrameP),
    
    #frame_params{grid_frame_id     = GridParentId,
		  grid_frame_width  = GridParentWidth,
		  grid_frame_height = GridParentHeight} = NewFrameP,

    PgPid ! #pg_init_grid{sender     = self(),
			  parent_id  = GridParentId,
			  width      = GridParentWidth,
			  height     = GridParentHeight,
			  xpos       = ?VBTN_WIDTH - 1,
			  ypos       = ?KEY_MARK_AREA_HEIGHT + ?HBTN_HEIGHT - 1,
			  nof_rows   = ?NOF_GRIDROWS,
			  row_height = ?ROW_HEIGHT
			 },


    receive
	#pg_col_info{first_col_shown     = FirstColShown, 
		     width_of_cols_shown = ColsShown,
		     nof_rows_shown      = NofRowsShown} ->
	    
	    PbPid ! #pb_init_btns{sender          = self(),
				  parent_id       = GridParentId,
				  parent_width    = GridParentWidth,
				  parent_height   = GridParentHeight,
				  ypos            = ?KEY_MARK_AREA_HEIGHT,
				  hbtn_height     = ?HBTN_HEIGHT,
				  resbtn_width    = ?RESBTN_WIDTH,
				  vbtn_width      = ?VBTN_WIDTH,
				  nof_rows        = ?NOF_GRIDROWS,
				  row_height      = ?ROW_HEIGHT,
				  first_col_shown = FirstColShown,
				  cols_shown      = ColsShown
				 },

	    NewScaleP = tv_pd_scale:init_scale(NewFrameP, ScaleP),	    

	    NewToolP  = init_toolbar(NewFrameP, ToolP),

	    ProcVars#process_variables{window_id       = WindowId,
				       window_width    = WindowWidth,
				       window_height   = WindowHeight,
				       first_col_shown = FirstColShown,
				       nof_rows_shown  = NofRowsShown,
				       cols_shown      = ColsShown,
				       frame_params    = NewFrameP,
				       scale_params    = NewScaleP,
				       toolbar_params  = NewToolP
				      }
    end.
    





%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


resize_display(NewWinW, NewWinH, ProcVars) ->
    #process_variables{pg_pid          = PgPid,
		       pb_pid          = PbPid,
		       color_list      = ColorList,
		       first_row_shown = FirstRowShown,
		       frame_params    = FrameP,
		       scale_params    = ScaleP,
		       toolbar_params  = ToolP}  = ProcVars,
    
    NewFrameP = tv_pd_frames:resize_display_frames(NewWinW, NewWinH, FrameP),
    
    #frame_params{grid_frame_width  = GridParentWidth,
		  grid_frame_height = GridParentHeight} = NewFrameP,
    
    PgPid ! #pg_resize_grid{sender = self(),
			    width  = GridParentWidth,
			    height = GridParentHeight
			   },

    receive
	#pg_col_info{first_col_shown     = FirstColShown, 
		     width_of_cols_shown = ColsShown,
		     nof_rows_shown      = NofRowsShown} ->
	    
	    PbPid ! #pb_update_hbtns{sender          = self(),
				     parent_width    = GridParentWidth,
				     parent_height   = GridParentHeight,
				     first_col_shown = FirstColShown,
				     cols_shown      = ColsShown
				    },

	    PbPid ! #pb_update_vbtns{sender           = self(),
				     color_list       = ColorList,
				     first_row_shown  = FirstRowShown,
				     nof_rows_shown   = NofRowsShown,
				     blinking_enabled = false
				    },

	    NewScaleP = tv_pd_scale:resize_scale(NewFrameP, ScaleP),

	    NewToolP = resize_toolbar(NewFrameP, ToolP),
	    
	    ProcVars#process_variables{window_width    = NewWinW,
				       window_height   = NewWinH,
				       first_col_shown = FirstColShown,
				       nof_rows_shown  = NofRowsShown,
				       cols_shown      = ColsShown,
				       frame_params    = NewFrameP,
				       scale_params    = NewScaleP,
				       toolbar_params  = NewToolP
				      }
    end.
    








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


resize_column(RealCol, VirtualCol, Xdiff, ProcVars) ->
    #process_variables{pg_pid = PgPid, 
		       pb_pid = PbPid,
		       frame_params = FrameP} = ProcVars,
    
    PgPid ! #pg_resize_grid_col{sender         = self(),
				real_col_no    = RealCol,
				virtual_col_no = VirtualCol,
				xdiff          = Xdiff
			       },
    
    #frame_params{grid_frame_width  = GridFrameWidth,
		  grid_frame_height = GridFrameHeight} = FrameP,
    receive
	#pg_col_info{first_col_shown     = FirstColShown,
		     width_of_cols_shown = ColsShown,
		     nof_rows_shown      = NofRowsShown} ->
	    
	    PbPid ! #pb_update_hbtns{parent_width    = GridFrameWidth,
				     parent_height   = GridFrameHeight,
				     first_col_shown = FirstColShown,
				     cols_shown      = ColsShown
				    },

	    ProcVars#process_variables{first_col_shown = FirstColShown,
				       nof_rows_shown  = NofRowsShown,
				       cols_shown      = ColsShown
				      }
    end.
    








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


display_data(Pos, Range, MaxValue, List, KeyList, MaxElemSize, MarkedRowData,ProcVars) ->
    #process_variables{master_pid     = PcPid,
		       pg_pid         = PgPid,
		       pb_pid         = PbPid,
		       sorting_on     = SortingOn,
		       nof_rows_shown = NofRowsShown,
		       scale_params   = ScaleP,
		       toolbar_params = ToolP, 
		       mark_params    = MarkP}  = ProcVars,

    {DataList, ColorList} = split_dblist(List, [], []),

    NewMarkP = update_marks(SortingOn, DataList, MarkedRowData, Pos, NofRowsShown,
			    PcPid, PgPid, ToolP, MarkP),

    PgPid ! #pg_data{sender = self(),
		     data   = DataList,
		     first_row_shown = Pos
		    },

    PbPid ! #pb_update_vbtns{sender           = self(),
			     color_list       = ColorList,
			     first_row_shown  = Pos,
			     nof_rows_shown   = NofRowsShown,
			     blinking_enabled = false
			    },

    PbPid ! #pb_key_info{sender       = self(),
			 list_of_keys = KeyList
			},

       % May be new number of elements in the total list!
    ?SCALE_FUNC_FILE:set_scale_range(vscale, Range, ScaleP),
       % May be new vertical scale position required!
    NewScaleP = ?SCALE_FUNC_FILE:set_scale_pos(vscale, Pos, ScaleP),
       % May be new maximum size of elements!
    ?SCALE_FUNC_FILE:set_scale_range(hscale, {1, MaxElemSize}, NewScaleP),
    
    ProcVars#process_variables{data_list       = DataList,
			       color_list      = ColorList,
			       first_row_shown = Pos,
			       scale_params    = NewScaleP,
			       mark_params     = NewMarkP
			      }.








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


scroll_vertically(MouseBtn, ProcVars) ->
    #process_variables{pg_pid       = PgPid,
		       pb_pid       = PbPid,
		       scale_params = ScaleP} = ProcVars,
    
    OldScalePos = ScaleP#scale_params.vscale_pos,
    NewScalePos  = get_new_scalepos(MouseBtn, OldScalePos),
    
    case NewScalePos of
	OldScalePos ->
	    ProcVars;
	NewValue ->
	    perform_vertical_scroll(NewValue, ProcVars)	    
    end.









%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


scroll_horizontally(MouseBtn, ProcVars) ->
    #process_variables{scale_params = ScaleP} = ProcVars,
    
    OldScalePos = ScaleP#scale_params.hscale_pos,
    NewScalePos  = get_new_scalepos(MouseBtn, OldScalePos),
    
    case NewScalePos of
	OldScalePos ->
	    ProcVars;
	NewValue ->
	    perform_horizontal_scroll(NewValue, ProcVars)
    end.











%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


perform_vertical_scroll(NewScalePos, ProcVars) ->
    #process_variables{master_pid   = MasterPid,
		       scale_params = ScaleP, 
		       mark_params  = MarkP} = ProcVars,
    
    MasterPid ! #pc_data_req{sender = self(),
			     element = NewScalePos,
			     nof_elements = ?NOF_GRIDROWS},
    
       % Since the order of click/buttonrelease messages isn't
       % precise, set the scale to the returned pos (may otherwise 
       % differ one unit).
    NewScaleP = ?SCALE_FUNC_FILE:set_scale_pos(vscale, 
					       NewScalePos, 
					       ScaleP),

    ProcVars#process_variables{scale_params = NewScaleP}.









%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


perform_horizontal_scroll(NewScalePos, ProcVars) ->
    #process_variables{pg_pid       = PgPid,
		       pb_pid       = PbPid,
		       frame_params = FrameP,
		       scale_params = ScaleP} = ProcVars,
    
       % Since the order of click/buttonrelease messages isn't
       % precise, set the scale to the returned pos (may otherwise 
       % differ one unit).
    NewScaleP = ?SCALE_FUNC_FILE:set_scale_pos(hscale, 
					       NewScalePos, 
					       ScaleP),

    PgPid ! #pg_horizontal_scroll{sender = self(),
				  leftmost_virtual_col = NewScalePos
				 },

    #frame_params{grid_frame_width  = GridFrameWidth,
		  grid_frame_height = GridFrameHeight} = FrameP,
    receive
	#pg_col_info{first_col_shown     = FirstColShown,
		     width_of_cols_shown = ColsShown,
		     nof_rows_shown      = NofRowsShown} ->
	    
	    PbPid ! #pb_update_hbtns{parent_width    = GridFrameWidth,
				     parent_height   = GridFrameHeight,
				     first_col_shown = FirstColShown,
				     cols_shown      = ColsShown
				    },
	    
	    ProcVars#process_variables{first_col_shown = FirstColShown,
				       cols_shown      = ColsShown,
				       nof_rows_shown  = NofRowsShown,
				       scale_params    = NewScaleP
				      }
    end.









%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


marked_cell(true, VirtualCol, RealRow, VirtualRow, ProcVars) ->
    #process_variables{master_pid     = MasterPid,
		       data_list      = DataList,
		       mark_params    = MarkP,
		       toolbar_params = ToolP} = ProcVars,
    
    DataElement = get_data_element(cell, DataList, RealRow, VirtualCol),
    update_toolbar_label(DataElement, ToolP, VirtualRow, VirtualCol),

    MasterPid ! #pc_marked_row{sender = self(),
			       row_no = VirtualRow
			      },

    NewMarkP = MarkP#mark_params{cell_col_no    = VirtualCol,
				 row_no         = RealRow,
				 virtual_row_no = VirtualRow
				},
    ProcVars#process_variables{mark_params = NewMarkP
			      };
marked_cell(false, VirtualCol, RealRow, VirtualRow, ProcVars) ->
    #process_variables{master_pid      = MasterPid, 
		       pb_pid          = PbPid,
		       mark_params     = MarkP} = ProcVars,

    PbPid ! #pb_remove_marks{sender           = self()},

    update_toolbar_label(notext, ProcVars#process_variables.toolbar_params, 
			 VirtualRow, VirtualCol),

    MasterPid ! #pc_marked_row{sender = self(),
			       row_no = VirtualRow
			      },

    NewMarkP = MarkP#mark_params{cell_col_no    = undefined,
				 row_no         = undefined,
				 virtual_row_no = undefined
				},
    ProcVars#process_variables{mark_params = NewMarkP
			      }.
				  







%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


update_toolbar_label(notext, ToolP, _VirtualRowNo, _VirtualColNo) ->
    #toolbar_params{fg_label_id = FgLblId,
		    editor_id   = EdId}  = ToolP,
    gs:config(FgLblId, [{label, {text, ""}}]),
    update_toolbar_editor(EdId, notext);
update_toolbar_label({DataToShow}, ToolP, VirtualRowNo, VirtualColNo) ->
    #toolbar_params{fg_label_id = FgLblId,
		    editor_id   = EdId}  = ToolP,

    case VirtualRowNo of
	undefined ->
	    done;  %% No row - nothing can possibly be marked!
	_AnyRow ->
	    RowStr  = " R" ++ integer_to_list(VirtualRowNo),
	    DataStr = lists:flatten(io_lib:write(DataToShow)),
	    Str = case VirtualColNo of
		      undefined ->
			  RowStr ++ ":  " ++ DataStr;
		      _AnyCol ->
			  RowStr ++ "xC" ++ integer_to_list(VirtualColNo) ++ ":  " ++ DataStr
		  end,
	    gs:config(FgLblId, [{label, {text,Str}}]),
	    update_toolbar_editor(EdId, {DataToShow})
    end.








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


get_data_element(row, DataList, RowNo, _VirtualCol) ->
    if
	length(DataList) < RowNo ->
	    notext;
	true ->
	    {lists:nth(RowNo, DataList)}
    end;
get_data_element(cell, DataList, RowNo, ColNo) ->
       % It's the responsibility of pg to ensure that there is a data item
       % for the cell marked, meaning we don't *have* to check the length of
       % the data items. However, since we in the future may want to edit
       % even empty cells, we check it!
    if
	length(DataList) < RowNo ->
	    notext;
	true ->
	    DataItem = lists:nth(RowNo, DataList),
	    if 
		tuple(DataItem) ->
		    if size(DataItem) < ColNo ->
			    notext;
		       true ->
			    {element(ColNo, DataItem)}
		    end;
		true ->
		    {DataItem}
	    end
    end.








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


show_toolbar_editor(ProcVars) ->
    #process_variables{frame_params   = FrameP,
		       toolbar_params = ToolP} = ProcVars,
    
    #frame_params{toolbar_frame_width  = TWidth,
		  toolbar_frame_height = THeight} = FrameP,
    
    #toolbar_params{editor_frame_id = EdFrameId} = ToolP,
    
    Xpos = 0,
    Ypos = THeight - 8,
    gs:config(EdFrameId, [{x, Xpos}, 
			  {y, Ypos}
			 ]),
    ProcVars.









%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


hide_toolbar_editor(ProcVars) ->
    #process_variables{toolbar_params = ToolP} = ProcVars,
    
    #toolbar_params{editor_frame_id = EdFrameId} = ToolP,
    
    Xpos = 0,
    Ypos = (-1) * gs:read(EdFrameId, height) - 50,
    gs:config(EdFrameId, [{x, Xpos}, 
			  {y, Ypos}
			 ]),
    ProcVars.








%%%********************************************************************
%%% INTERNAL FUNCTIONS
%%%********************************************************************






%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


update_toolbar_editor(EdId, notext) ->
    gs:config(EdId, [{enable, true}]),
    gs:config(EdId, [clear]),
    gs:config(EdId, [{enable, false}]);
update_toolbar_editor(EdId, {DataToShow}) ->
    Str = io_lib:format("~n~p~n", [DataToShow]),
    gs:config(EdId, [{enable, true}]),
    gs:config(EdId, [clear]),
    gs:config(EdId, [{overwrite, {insert, Str}}]),
    gs:config(EdId, [{enable, false}]).
    








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


update_marks(true, DataList, 
	     MarkedRowData, Pos, NofRowsShown, PcPid, PgPid, ToolP, MarkP) ->
    PgPid ! #pg_remove_marks{sender = self()},
    update_toolbar_label(notext, ToolP, undefined, undefined),
    PcPid ! #pc_marked_row{sender = self(),
			   row_no = undefined
			  },
    MarkP#mark_params{cell_col_no    = undefined,
		      row_no         = undefined,
		      virtual_row_no = undefined
		     };
update_marks(false, DataList, 
	     MarkedRowData, Pos, NofRowsShown, PcPid, PgPid, ToolP, MarkP) ->
    #mark_params{cell_col_no    = CellColNo,
		 row_no         = RowNo,
		 virtual_row_no = VirtualRowNo}  = MarkP,
    
       % Marked row data contains the color also!
    {RowData, _RowColor} = split_dblist(MarkedRowData, [], []),

    DataElement = choose_data_to_show(VirtualRowNo, CellColNo, RowData, 
				      DataList, Pos),

    update_toolbar_label(DataElement, ToolP, VirtualRowNo, CellColNo),
    MarkP.
			  

				  
    






%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


choose_data_to_show(VirtualRowNo, undefined, RowData, DataList, Pos) when VirtualRowNo >= Pos, VirtualRowNo =< (Pos + length(DataList) - 1) ->
    get_data_element(row, DataList, VirtualRowNo - Pos + 1, undefined);
choose_data_to_show(VirtualRowNo, undefined, RowData, DataList, Pos) ->
    get_data_element(row, RowData, 1, undefined);
choose_data_to_show(VirtualRowNo, CellColNo, RowData, DataList, Pos) when VirtualRowNo >= Pos, VirtualRowNo =< (Pos + length(DataList) - 1) ->
    get_data_element(cell, DataList, VirtualRowNo - Pos + 1, CellColNo);
choose_data_to_show(VirtualRowNo, CellColNo, RowData, DataList, Pos) ->
    get_data_element(cell, RowData, 1, CellColNo).
    










%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================




get_new_scalepos(Btn, LastScalePos) ->
    receive
	{gs, _Id, click, _Data, [NewScalePos | _T]} ->
	    get_new_scalepos(Btn, NewScalePos);
	
	{gs, _Id, buttonrelease, _Data, [Btn | _T]} ->
	    LastScalePos;
	
	{gs, _Id, buttonrelease, _Data, _Args} ->
	    get_new_scalepos(Btn, LastScalePos);
	
	{gs, _Id, buttonpress, _Data, _Args} ->
	    get_new_scalepos(Btn, LastScalePos)
    
    end.










%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


split_dblist([], DataAcc, ColorAcc) -> 
    {lists:reverse(DataAcc), lists:reverse(ColorAcc)};
split_dblist([{Data, Color} | Tail], DataAcc, ColorAcc) ->
    split_dblist(Tail, [Data | DataAcc], [Color | ColorAcc]).


    












%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


init_toolbar(FrameP, ToolP) ->
    #frame_params{display_id           = DispId,
		  toolbar_frame_id     = TId,
		  toolbar_frame_width  = TWidth,
		  toolbar_frame_height = THeight} = FrameP,

    NewToolP = init_toolbar_btns(TId, ToolP),
    {BgLabelId, FgLabelId, BtnId} = init_toolbar_label(TId, TWidth, THeight),
    
    PopUpFrame = gs:frame(TId, [{width, 80},
				{height, 20},
				{x, 0},
				{y, -30},
				{bg, {0, 0, 0}}
			       ]),
    
    PopUpLabel = gs:label(PopUpFrame, [{width, 78}, 
				       {height, 18}, 
				       {bg, {255,255,190}}, 
				       {x,1}, 
				       {y,1}, 
				       {align, center},
				       {label, {text,""}}, 
				       {font,{screen,12}}]),
    
    {EditorFrameId, EditorId} = init_toolbar_editor(DispId, TWidth, THeight),

    NewToolP#toolbar_params{parent_id       = TId,
			    bg_label_id     = BgLabelId,
			    fg_label_id     = FgLabelId,
			    label_btn_id    = BtnId,
			    pop_up_frame_id = PopUpFrame,
			    pop_up_label_id = PopUpLabel,
			    editor_frame_id = EditorFrameId,
			    editor_id       = EditorId
			}.







%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


init_toolbar_btns(TId, ToolP) ->
    PicDir = code:priv_dir(tv),
       % Toolbar btns are 23x23, the bitmaps are 20x20.
    B1 = create_one_toolbar_btn(TId, 1, PicDir ++ "/open.xbm",
				{toolbar, open_table, "Open Table"}),
    B3 = create_one_toolbar_btn(TId, 3, PicDir ++ "/search.xbm",
				{toolbar, search_object, "Search Object"}),
    B5 = create_one_toolbar_btn(TId, 5, PicDir ++ "/sort.xbm",
				{toolbar, sort_rising_order, "Sort Rising Order"}),
    B6 = create_one_toolbar_btn(TId, 6, PicDir ++ "/no_sort.xbm",
				{toolbar, no_sorting,"No Sorting"}),
    B7 = create_one_toolbar_btn(TId, 7, PicDir ++ "/sort_reverse.xbm",
				{toolbar, sort_falling_order,"Sort Falling Order"}),
    B9 = create_one_toolbar_btn(TId, 9, PicDir ++ "/poll.xbm",
				{toolbar, poll_table,"Poll Table"}),
    B11 = create_one_toolbar_btn(TId, 11, PicDir ++ "/info.xbm",
				{toolbar, table_info,"Table Info"}),
    B13 = create_one_toolbar_btn(TId, 13, PicDir ++ "/help.xbm",
				 {toolbar, help_button, "Help"}),
    ToolP.
    









%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


create_one_toolbar_btn(ParentId, N, Image, Data) ->
    BtnWidth  = 25,
    BtnHeight = 25,
    StartXpos = 0,
    BtnXpos   = StartXpos + ((N - 1) * BtnWidth),
    BtnYpos   = 2,
    BgColor   = ?DEFAULT_BG_COLOR,
    FgColor   = {178,34,34},       % Firebrick

    gs:button(ParentId, [{width, BtnWidth},
			 {height, BtnHeight}, 
			 {x, BtnXpos}, 
			 {y, BtnYpos},
			 {enter, true},
			 {leave, true},
			 {label, {image, Image}}, 
			 {data, Data},
			 {fg, FgColor},
			 {bg, BgColor}
			]).
    








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


resize_toolbar(FrameP, ToolP) ->
    #frame_params{toolbar_frame_width  = TWidth,
		  toolbar_frame_height = THeight} = FrameP,
    
    #toolbar_params{bg_label_id     = BgId,
		    fg_label_id     = FgId,
		    label_btn_id    = BtnId,
		    editor_frame_id = FrId,
		    editor_id       = EdId} = ToolP,

    resize_toolbar_label(BgId, FgId, BtnId, TWidth, THeight),
    resize_toolbar_editor(FrId, EdId, TWidth, THeight),
    ToolP.











%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


init_toolbar_label(ParentId, ParentWidth, ParentHeight) ->
    {BgWidth, BgHeight, BgXpos, BgYpos, FgWidth, FgHeight, FgXpos, FgYpos, BtnWidth,
     BtnHeight, BtnXpos, BtnYpos} = 
	get_toolbar_label_coords(ParentWidth, ParentHeight),

    BgId = gs:label(ParentId, [{width, BgWidth},
			       {height, BgHeight},
			       {x, BgXpos},
			       {y, BgYpos},
			       {bg, {0, 0, 0}},
			       {fg, {0, 0, 0}}
			      ]),

    FgId = gs:label(ParentId, [{width, FgWidth},
			       {height, FgHeight},
			       {x, FgXpos},
			       {y, FgYpos},
			       {bg, {255, 255, 255}},
			       {fg, {0, 0, 0}},
			       {align, w},
			       {cursor, text}
			      ]),

    PicDir = code:priv_dir(tv),
    BtnId = gs:button(ParentId, [{width, BtnWidth},
				 {height, BtnHeight},
				 {x, BtnXpos},
				 {y, BtnYpos},
				 {bg, ?DEFAULT_BG_COLOR},
				 {fg, {0, 0, 0}},
				 {label, {image, PicDir ++ "/more.xbm"}},
				 {data, {labelbtn, pop_up}}
				]),

    {BgId, FgId, BtnId}.
    







%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


init_toolbar_editor(DispId, TWidth, THeight) ->
    {BgWidth, BgHeight, BgXpos, BgYpos, Width, Height, Xpos, Ypos} = 
	get_toolbar_editor_coords(TWidth, THeight),
    
    EditorFrame = gs:frame(DispId, [{width, BgWidth},
				    {height, BgHeight},
				    {x, BgXpos},
				    {y, BgYpos},
				    {bg, {0, 0, 0}}
				   ]),

    Editor = gs:editor(EditorFrame, [{width, Width},
				     {height, Height},
				     {x, Xpos},
				     {y, Ypos},
				     {vscroll, right},
				     {wrap, word},
				     {bg, {255, 255, 255}},
				     {fg, {0, 0, 0}},
				     {enable, false}
				    ]),
    
    {EditorFrame, Editor}.
    









%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


get_toolbar_editor_coords(TWidth, THeight) ->
    BgWidth  = TWidth - 20,
    BgHeight = 200,
    BgXpos   = 0,
    BgYpos   = (-1) * BgHeight - 50,
    FgWidth  = BgWidth - 2,
    FgHeight = BgHeight - 2,
    FgXpos   = 1,
    FgYpos   = 1,

    {BgWidth, BgHeight, BgXpos, BgYpos, FgWidth, FgHeight, FgXpos, FgYpos}.











%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


resize_toolbar_editor(FrId, EdId, TWidth, THeight) ->
    {BgWidth, BgHeight, BgXpos, BgYpos, FgWidth, FgHeight, FgXpos, FgYpos} = 
	get_toolbar_editor_coords(TWidth, THeight),
    gs:config(FrId, [{width, BgWidth},
		     {height, BgHeight}
		    ]),

    gs:config(EdId, [{width, FgWidth},
		     {height, FgHeight}
		    ]).








%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


resize_toolbar_label(BgId, FgId, BtnId, ParentWidth, ParentHeight) ->
    {BgWidth, BgHeight, BgXpos, BgYpos, FgWidth, FgHeight, FgXpos, FgYpos, BtnWidth,
     BtnHeight, BtnXpos, BtnYpos} = 
	get_toolbar_label_coords(ParentWidth, ParentHeight),

    gs:config(BgId, [{width, BgWidth},
		     {height, BgHeight}
		    ]),

    gs:config(BtnId, [{x, BtnXpos},
		      {y, BtnYpos}
		     ]),

    gs:config(FgId, [{width, FgWidth},
		     {height, FgHeight}
		    ]).    






%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


get_toolbar_label_coords(ParentWidth, ParentHeight) ->
    BtnWidth  = 19,
    BgWidth   = ParentWidth,
%    BgWidth  = ParentWidth - ?VBTN_WIDTH - ?VSCALE_WIDTH - 2,
    BgHeight  = 26,
    BgXpos    = 0,
    BgYpos    = ParentHeight - BgHeight - 8,
    FgHeight  = BgHeight - 2,
    FgWidth   = BgWidth - BtnWidth - 3,
    FgXpos    = BgXpos + 1,
    FgYpos    = BgYpos + 1,
    BtnHeight = BgHeight - 2,
    BtnXpos   = FgWidth + 2,
    BtnYpos   = BgYpos + 1,

    {BgWidth, BgHeight, BgXpos, BgYpos, FgWidth, FgHeight, FgXpos, FgYpos, BtnWidth,
    BtnHeight, BtnXpos, BtnYpos}.
    










