%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%*********************************************************************
%%%
%%%   Filename:         tv_tab_dialog.erl
%%%
%%%   Author:           EPK/TE  Fredrik Gustafson 
%%% 
%%%   SCCS Delta ID:    1.3
%%%
%%%   SCCS Delta Date:  97/06/12
%%%
%%%   SCCS ID:          @(#)  tv_tab_dialog.erl  1.3  97/06/12
%%%
%%%   Description:      Code for the "Open Table" dialog with the user.
%%%
%%%*********************************************************************
%%%
%%%   COPYRIGHT ERICSSON TELECOM AB 1998
%%%
%%%   THE COPYRIGHT TO THE COMPUTER PROGRAM HEREIN IS THE
%%%   PROPERTY OF ERICSSON TELECOM AB, SWEDEN. THE PROGRAM
%%%   MAY BE USED AND/OR COPIED ONLY WITH THE WRITTEN PERMISSION
%%%   FROM ERICSSON TELECOM AB, OR IN ACCORDANCE WITH THE TERMS
%%%   AND CONDITIONS STIPULATED IN THE AGREEMENT/CONTRACT UNDER
%%%   WHICH THE PROGRAM HAS BEEN SUPPLIED.
%%%
%%%*********************************************************************
%%%
%%%   Revision information:
%%%
%%%   Date      Name                       Description
%%%   ------    -----------------------    -----------
%%%   970427    EPK/TE  F Gustafson        First version.
%%%   970526    EPK/TE  F Gustafson        Added that window is always on top, 
%%%                                        and that some signals are stored,
%%%                                        e.g., window resizing.
%%%   971210    EPK/TE  F Gustafson        Added possibility to choose distributed
%%%                                        tables.
%%%
%%%*********************************************************************



-module(tv_tab_dialog).
-author('fredrik@erlang.ericsson.se').
-vsn('$Revision: /main/release/free/1').
-copyright('Copyright (C) 1991-98 Ericsson Telecom AB').



-export([start/2, init/3]).

-export([get_ets_tables/0]).



-include("tv_int_msg.hrl").


-record(tab_info, {tab_id,
		   name,
		   owner,
		   protection,
		   name_no,
		   text
		  }).




-define(MIDDLE, 270).
-define(WINDOW_WIDTH, 540).     % 2 * MIDDLE
-define(WINDOW_HEIGHT, 410).
-define(DEFAULT_BG_COLOR, {217, 217, 217}).



%% TABLES_TO_HIDE shall contain both Mnesia and ETS tables that we want to hide.  :-)

-define(SYSTEM_TABLES, [queue,
			gvar, 
			tid_locks, 
			locks, 
			sticky,
			clist, 
			lmcounter,
			dets_registry,
			index,
			gtk_db,
			int_db,
			inet_cache,
			inet_db,
			inet_hosts,
			sys_dist,
			coast_internal_data_table
		       ]).


-define(MNESIA_TABLES, [schema,
			mnesia_decision,
			mnesia_transient_decision,
			mnesia_clist
		       ]).


-define(TABLES_TO_HIDE, ?SYSTEM_TABLES ++ ?MNESIA_TABLES).



%%%*********************************************************************
%%% EXTERNAL FUNCTIONS
%%%*********************************************************************




%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


start(KindOfTable, CurrNode) ->
    ProcPid = spawn_link(?MODULE, init, [self(), KindOfTable, CurrNode]),
    receive_answer(ProcPid).





%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


init(Pid, KindOfTable, CurrNode) ->
    process_flag(trap_exit, true),
    create_window(KindOfTable),
    NewCurrNode = update_node_listbox(CurrNode),
    update_table_listboxes(NewCurrNode, KindOfTable),
    gs:config(win, [{map,true}, {cursor,arrow}]),
    browser_loop(Pid, KindOfTable, NewCurrNode, no_table).






get_ets_tables() ->
    Tables = ets:all(),
    TablesWInfo = get_ets_table_info(Tables, ?TABLES_TO_HIDE ++ current_mnesia_tables(), []),
    enumerate_tables(TablesWInfo, {}, 0).
    



%%%*********************************************************************
%%% INTERNAL FUNCTIONS
%%%*********************************************************************




analyze_error(Cause) ->
    case Cause of
	{badrpc, {'EXIT', {badarg,Reason}}} ->
	    done;  %% Table has ceased to exist.
	{'EXIT', {badarg, {ets,local_info,Args}}} ->  
	    done;
	
	{badrpc, nodedown} ->
            handle_error(nodedown);
	{'EXIT', nodedown} ->
            handle_error(nodedown);

        {'EXIT', {aborted, {node_not_running,_Node}}} ->
            handle_error(mnesia_not_started);
        {'EXIT', {'EXIT', {aborted, {node_not_running,_Node}}}} ->
            handle_error(mnesia_not_started);
        {badrpc, {'EXIT', {aborted, {node_not_running,_Node}}}} ->
            handle_error(mnesia_not_started);
	{'EXIT', {undef, {mnesia,Fcn,Args}}} ->
	    handle_error(mnesia_not_started);

        {'EXIT', Reason} ->
            handle_error({unexpected_error,Reason});
        Error when tuple(Error) ->
            handle_error({unexpected_error,Error})
    end.
    
	    


handle_error(mnesia_not_started) ->
    show_info("Mnesia not started!", [beep]);
handle_error(no_table) ->
    gs:config(win, [beep]),
    tool_utils:notify(win, ["The table doesn't exist anymore!"]);
handle_error(nodedown) ->
    gs:config(win, [beep]),
    tool_utils:notify(win, ["The selected node is down!"]),
    self() ! nodedown;
handle_error({unexpected_error,Cause}) ->
    io:format("Unexpected error:  ~p~n", [Cause]),
    gs:config(win, [beep]).





update_table_listboxes(Node, KindOfTable) ->
    gs:config(win, [{cursor,busy}]),
    case catch update_table_listboxes2(Node, KindOfTable) of
	ok ->
	    done;
	Other ->
	    analyze_error(Other),
	    gs:config(readable_listbox, [{data,[]}, {items,[]}]),
	    gs:config(unreadable_listbox, [{items, []}])
    end,
    gs:config(win, [{cursor,arrow}]).





update_table_listboxes2(Node, KindOfTable) ->
    Tables        = get_table_list(Node, KindOfTable),
    {RData,UData} = split_in_readable_unreadable(KindOfTable, Tables),
    gs:config(readable_listbox, [{data,RData}, 
				 {items,tables_as_text(KindOfTable, RData)}
				]),
    gs:config(unreadable_listbox, [{items,tables_as_text(KindOfTable, UData)}
				  ]),
    ok.
    


    




get_node_lists(CurrNode) ->
    NodeDataList = lists:sort([node()] ++ nodes()),
    NodeTextList = lists:map(fun(Item) ->
					 " " ++ atom_to_list(Item)
				 end, 
				 NodeDataList),

    %% It *may* be possible that CurrNode has disappeared!
    %% If this is the case, use the node where TV resides
    %% as new current node.
    %% This also covers the case when our own node (or some
    %% other node) suddenly goes distributed.

    NewCurrNode = case lists:member(CurrNode, NodeDataList) of
		      true ->
			  CurrNode;
		      false ->
			  node()
		  end,
    
       %% Now get the index that shall be marked in the node listbox.
       %% Remember that the first item has number 0 (zero)!
    NodeMarkIndex = get_node_mark_index(NewCurrNode, NodeDataList, 0),
    
    {NewCurrNode, NodeDataList, NodeTextList, NodeMarkIndex}.
    



%% We know that CurrNode is *somewhere* in the list, since we have checked.
%% If the original CurrNode wasn't there, then we are using node() instead,
%% which definitely is in the list. (node() may have gone distributed in the 
%% meantime, but it *IS* in the list!)  :-)

get_node_mark_index(CurrNode, [H | T], Acc) when CurrNode /= H ->
    get_node_mark_index(CurrNode, T, Acc + 1);
get_node_mark_index(CurrNode, [CurrNode | T], Acc) ->
    Acc.  %% Acc tells the index of the current head.  :-)
    




%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


receive_answer(ProcPid) ->
    receive_answer(ProcPid, undefined, undefined, undefined, undefined).







%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


receive_answer(ProcPid, DataReqMsg, WinConfMsg, MarkedRowMsg, SubsetMsg) ->
    receive Msg ->
	    case Msg of
		{browser, ProcPid, cancel} ->
		    PcPid = self(),
		    PcPid ! DataReqMsg,
		    PcPid ! WinConfMsg,
		    PcPid ! MarkedRowMsg,
		    PcPid ! SubsetMsg,
		    cancel;

		{browser, ProcPid, 
		 {ok, Node, TabId, KindOfTable, TabName, TabNumber}} ->
		       %% No need to manipulate old table data if a new table 
		       %% has been chosen! Just resize window, if necessary!
		    self() ! WinConfMsg,
		    {Node, TabId, KindOfTable, TabName, TabNumber};

		#pc_data_req{} ->
		    receive_answer(ProcPid, Msg, WinConfMsg, MarkedRowMsg, SubsetMsg);

		#pc_win_conf{} ->
		    receive_answer(ProcPid, DataReqMsg, Msg, MarkedRowMsg, SubsetMsg);

		#pc_marked_row{} ->
		    receive_answer(ProcPid, DataReqMsg, WinConfMsg, Msg, SubsetMsg);

		#dbs_subset{} ->
		    receive_answer(ProcPid, DataReqMsg, WinConfMsg, MarkedRowMsg, Msg);

		#pc_menu_msg{data = exit_button} ->
		    self() ! Msg,
		    cancel;

		#pc_set_sorting_mode{sender = Sender} ->
		    Sender ! #pd_ignore{sender = self()},
		    ProcPid ! raise_and_beep,
		    receive_answer(ProcPid, DataReqMsg, WinConfMsg, MarkedRowMsg, SubsetMsg);

		{'EXIT', Sender, Reason} ->
		    self() ! Msg,
		    cancel;

		undefined ->
		    receive_answer(ProcPid, DataReqMsg, WinConfMsg, MarkedRowMsg, SubsetMsg);
		
		Other ->
		    ProcPid ! raise_and_beep,
		    receive_answer(ProcPid, DataReqMsg, WinConfMsg, MarkedRowMsg, SubsetMsg)
	    end
    end.
    






%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


browser_loop(Pid, KindOfTable, CurrNode, TableChosenInfo) ->
       %% TableChosenInfo = {TableId, TableText}
    receive

	{gs, ets_btn, click, _, _} when KindOfTable == ets ->
	    show_info(""),
	    browser_loop(Pid, ets, CurrNode, no_table);


	{gs, ets_btn, click, _, _} ->
	    config_listbox_labels(ets),
	    update_table_listboxes(CurrNode, ets),
	    browser_loop(Pid, ets, CurrNode, no_table);


	{gs, mnesia_btn, click, _, _} when KindOfTable == mnesia ->
	    show_info(""),
	    browser_loop(Pid, mnesia, CurrNode, no_table);


	{gs, mnesia_btn, click, _, _} ->
	    config_listbox_labels(mnesia),
	    update_table_listboxes(CurrNode, mnesia),
	    browser_loop(Pid, mnesia, CurrNode, no_table);

	
	{gs, node_listbox, click, Data, [Idx, Txt | _]} ->
	    show_info(""),
	    TmpCurrNode = lists:nth(Idx + 1, Data),
	    NewCurrNode = 
		case TmpCurrNode of 
		    CurrNode ->
			case net_adm:ping(CurrNode) of
			    pong ->
				CurrNode;
			    _Other ->
				handle_error(nodedown),
				NewNode = update_node_listbox(CurrNode),
				config_listbox_labels(KindOfTable),
				update_table_listboxes(NewNode, KindOfTable),
				NewNode
			end;
		    _Other ->
			config_listbox_labels(KindOfTable),
			update_table_listboxes(TmpCurrNode, KindOfTable),
			TmpCurrNode
		end,
	    browser_loop(Pid, KindOfTable, NewCurrNode, no_table);
	

	{gs, readable_listbox, click, Data, [Idx, Txt | _]} ->
	    show_info(""),
	    TableInfo = lists:nth(Idx + 1, Data),
	    {TabId, Name, Number} = extract_table_info(KindOfTable, 
						       TableInfo),
	    case check_table_alive(CurrNode, TabId, KindOfTable) of
		ok ->
		    browser_loop(Pid, KindOfTable, CurrNode, TableInfo);
		continue ->
		    browser_loop(Pid, KindOfTable, CurrNode, no_table)
	    end;


	{gs, readable_listbox, doubleclick, Data, [Idx, Txt | _]} ->
	    TableInfo = lists:nth(Idx + 1, Data),
	    {TabId, Name, Number} = extract_table_info(KindOfTable, TableInfo),
	    case check_table_alive(CurrNode, TabId, KindOfTable) of
		ok ->
		    show_info(""),
		    Pid ! {browser, self(), {ok, CurrNode, TabId, 
					     KindOfTable, Name, Number}};
		continue ->
		    browser_loop(Pid, KindOfTable, CurrNode, no_table)
	    end;


	{gs, unreadable_listbox, click, _, [_Idx, Txt | _]} ->
	    gs:config(unreadable_listbox, [{selection, clear}]),
	    gs:config(readable_listbox, [{selection, clear}]),
	    show_info(""),
	    browser_loop(Pid, KindOfTable, CurrNode, no_table);


	{gs, unreadable_btn, click, show, _} ->
	    gs:config(unreadable_frame, [{y, 58}]),
	    show_info(""),
	    gs:config(unreadable_btn, [{data, hide}]),
	    gs:config(readable_listbox, [{selection, clear}]),
	    browser_loop(Pid, KindOfTable, CurrNode, TableChosenInfo);


	{gs, unreadable_btn, click, hide, _} ->
	    gs:config(unreadable_frame, [{y, -308}]),
	    show_info(""),
	    gs:config(unreadable_btn, [{data, show}]),
	    browser_loop(Pid, KindOfTable, CurrNode, TableChosenInfo);


	{gs, open_btn, click, _, _} ->
	    {TabId, Name, Number} = extract_table_info(KindOfTable, 
						       TableChosenInfo),
	    case check_table_alive(CurrNode, TabId, KindOfTable) of
		ok ->
		    Pid ! {browser, self(), {ok, CurrNode, TabId, 
					     KindOfTable, Name, Number}};
		continue ->
		    browser_loop(Pid, KindOfTable, CurrNode, no_table)
	    end;



	{gs, rescan_btn, click, _, _} ->
	    NewCurrNode = update_node_listbox(CurrNode),
	    config_listbox_labels(KindOfTable),
	    update_table_listboxes(NewCurrNode, KindOfTable),
	    browser_loop(Pid, KindOfTable, NewCurrNode, no_table);


	
	{gs, cancel_btn, click, _, _} ->
	    show_info(""),
	    Pid ! {browser, self(), cancel};



	{gs, _, destroy, _, _} ->
	    Pid ! {browser, self(), cancel},
	    exit(normal);

	

	{gs, win, configure, _, _} ->
	    gs:config(win, [{width, ?WINDOW_WIDTH},
			    {height, ?WINDOW_HEIGHT}
			   ]),
	    browser_loop(Pid, KindOfTable, CurrNode, TableChosenInfo);


	raise_and_beep ->
	    gs:config(win, [raise,
			    beep
			   ]),
	    browser_loop(Pid, KindOfTable, CurrNode, TableChosenInfo);


	nodedown ->
	    show_info(""),
	    NewCurrNode = update_node_listbox(CurrNode),
	    case NewCurrNode of 
		CurrNode ->
		    done;
		_Other ->
		    config_listbox_labels(KindOfTable),
		    update_table_listboxes(NewCurrNode, KindOfTable)
	    end,
	    browser_loop(Pid, KindOfTable, NewCurrNode, no_table);


	{'EXIT', Sender, Reason} ->
	    Pid ! {browser, self(), cancel},
	    exit(normal);


	_Other ->
	    io:format("Open table dialog received message ~p ~n", [_Other]),
	    browser_loop(Pid, KindOfTable, CurrNode, TableChosenInfo)

    end.
			      


extract_table_info(_KindOfTable, no_table) ->
    {no_table, undefined, undefined};
extract_table_info(mnesia, TabId) ->
    {TabId, TabId, undefined};
extract_table_info(ets, TabRec) ->
    {TabRec#tab_info.tab_id, TabRec#tab_info.name, TabRec#tab_info.name_no}.



update_node_listbox(Node) ->
    {NewNode, NodeDataList, NodeTextList, MarkIndex} = get_node_lists(Node),
    gs:config(node_listbox, [{data, NodeDataList},
			     {items, NodeTextList},
			     {selection, MarkIndex}
			    ]),
    NewNode.





show_info(Text) ->
    show_info(Text, []).


show_info(Text, Opts) ->
    gs:config(info_label, [{label, {text,Text}}]),
    gs:config(info_label, Opts).






%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================

check_table_alive(Node, no_table, KindOfTable) ->
    show_info("No table chosen!", [beep]),
    continue;
check_table_alive(Node, TabId, KindOfTable) ->
    case catch check_table_alive2(KindOfTable, Node, TabId) of
	ok ->
	    ok;
	Other ->
	    handle_error(Other),
	    continue
    end.




check_table_alive2(mnesia, Node, TabId) ->
    case tv_mnesia_rpc:table_info(Node, TabId, access_mode) of
	_Anything ->  %% Errors are caught above, in check_table_alive/3!
	    ok
    end;
check_table_alive2(ets, Node, TabId) ->
    case tv_ets_rpc:info(Node, TabId, name) of
	_Anything ->
	    ok
    end.
	    
		    
    



%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


config_listbox_labels(ets) ->
    gs:config(readable_listbox_label, 
	      [{label, {text,"Readable tables (and owners):  "}}]),
    gs:config(unreadable_listbox_label, 
	      [{label, {text,"Unreadable tables (and owners):"}}]);
config_listbox_labels(mnesia) ->
    gs:config(readable_listbox_label, 
	      [{label, {text,"Readable tables:  "}}]),
    gs:config(unreadable_listbox_label, 
	      [{label, {text,"Unreadable tables:"}}]).
    
    





%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================




tables_as_text(ets, Tables) ->
    MaxLength = get_length_of_longest_table_name(Tables, 0),
    add_owners(Tables, MaxLength, []);
tables_as_text(mnesia, [Table | Tail]) ->
    [" " ++ atom_to_list(Table) | tables_as_text(mnesia, Tail)];
tables_as_text(mnesia, []) ->
    [].





get_length_of_longest_table_name([], MaxLength) ->
    MaxLength;
get_length_of_longest_table_name([TabRec | Rest], CurrMax) ->
    L = length(TabRec#tab_info.text),
    if 
	L > CurrMax ->
	    get_length_of_longest_table_name(Rest, L);
	true ->
	    get_length_of_longest_table_name(Rest, CurrMax)
    end.






add_owners([], _MaxLength, Acc) ->
    lists:reverse(Acc);
add_owners([TabRec | Tail], MaxLength, Acc) ->
    Text  = TabRec#tab_info.text ++ " ",
    Owner = TabRec#tab_info.owner,
    add_owners(Tail, 
	       MaxLength, 
	       [Text ++ 
		lists:duplicate(MaxLength-(length(Text)-2)," ") ++ 
		Owner 
		| Acc
	       ]).




%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


get_table_list('nonode@nohost', ets) ->
    get_ets_tables();
get_table_list(Node, ets) when Node == node() ->
    get_ets_tables();
get_table_list(Node, ets) when node() == 'nonode@nohost' ->
    get_ets_tables();
get_table_list(Node, ets) ->
    case rpc:block_call(Node, ?MODULE, get_ets_tables, []) of
	{badrpc, Reason} ->
	    throw({badrpc,Reason});
	Result ->
	    Result
    end;
get_table_list('nonode@nohost', mnesia) ->
    Tables = mnesia:system_info(tables),
    lists:sort(Tables -- ?TABLES_TO_HIDE);
get_table_list(Node, mnesia) when Node == node() ->
    Tables = mnesia:system_info(tables),
    lists:sort(Tables -- ?TABLES_TO_HIDE);
get_table_list(Node, mnesia) when node() == 'nonode@nohost' ->
    Tables = mnesia:system_info(tables),
    lists:sort(Tables -- ?TABLES_TO_HIDE);
get_table_list(Node, mnesia) ->
    case rpc:block_call(Node, mnesia, system_info, [tables]) of
	{badrpc,Reason} ->
	    throw({badrpc,Reason});
	Result ->
	    lists:sort(Result -- ?TABLES_TO_HIDE)
    end.


    

    
get_ets_table_info([], TablesToHide, Acc) ->
    lists:keysort(#tab_info.name, Acc);
get_ets_table_info([TabId | Tail], TablesToHide, Acc) ->
    case catch get_ets_name_owner_protection(TabId) of
	{'EXIT', Reason} ->  
	       %% Ignore tables ceasing to exist. 
	       %% Nodedown errors caught above!
	    get_ets_table_info(Tail, TablesToHide, Acc);
	{Name, Owner, Protection} ->
	    case lists:member(Name, TablesToHide) of
		true ->
		    get_ets_table_info(Tail, TablesToHide, Acc);
		false ->
		    get_ets_table_info(Tail, TablesToHide, 
				       [#tab_info{tab_id      = TabId,
						  name        = Name,
						  owner       = "(" ++ Owner ++ ")", 
						  protection  = Protection,
						  text	      = " " ++ atom_to_list(Name)
						 }  
					| Acc])
	    end
    end.




remove_tables_to_hide([{EtsName,Pid} | T], HiddenTables, Acc) ->
    case lists:member(EtsName, HiddenTables) of
	true ->
	    remove_tables_to_hide(T, HiddenTables, Acc);
	false ->
	    remove_tables_to_hide(T, HiddenTables, [{EtsName,Pid} | Acc])
    end;
remove_tables_to_hide([], HiddenTables, Acc) ->
    lists:reverse(Acc).
				  





enumerate_tables([TabRec | Tail],
		 PrevName,PrevNameNo) when TabRec#tab_info.name == PrevName ->
    NameNo = PrevNameNo + 1,
    NewTabRec = TabRec#tab_info{name_no = NameNo,
				text = TabRec#tab_info.text ++ "   #" ++ 
				          integer_to_list(NameNo)},
    [NewTabRec | enumerate_tables(Tail, PrevName, NameNo)];
enumerate_tables([TabRec | [H|T]], PrevName,
		 PrevNameNo) when TabRec#tab_info.name /= H#tab_info.name ->
    NewTabRec = TabRec#tab_info{name_no = undefined,
				text = TabRec#tab_info.text ++ "   "},
    [NewTabRec | enumerate_tables([H|T], TabRec#tab_info.name, 0)];
enumerate_tables([TabRec | [H|T]],PrevName,PrevNameNo) ->
    NewTabRec = TabRec#tab_info{name_no = 1,
				text = TabRec#tab_info.text ++ "   #1"},
    [NewTabRec | enumerate_tables([H|T], TabRec#tab_info.name, 1)];
enumerate_tables([TabRec],PrevName,PrevNameNo) ->
    NewTabRec = TabRec#tab_info{name_no = undefined,
				text    = TabRec#tab_info.text ++ "   "},
    [NewTabRec];
enumerate_tables([], _PrevName, _PrevNameNo) ->
    [].


    




	       
get_ets_name_owner_protection(TabId) ->
    Name     = ets:info(TabId, name),
    OwnerPid = ets:info(TabId, owner),
    Protect  = ets:info(TabId, protection),
    PName    = case lists:keysearch(registered_name,1,process_info(OwnerPid)) of
		   false ->
		       pid_to_list(OwnerPid);  %% We *have* the pid!
		   {value, {registered_name, ProcName}} ->
		       atom_to_list(ProcName)
	       end,
    {Name, PName, Protect}.

    




current_mnesia_tables() ->
    case catch get_table_list(node(), mnesia) of
	Result when list(Result) ->
	    Result;
	nodedown ->
	    handle_error(nodedown),
	    [];
	_Other ->
	    []
    end.
	    






%%======================================================================
%% Function:      
%%
%% Return Value:  
%%
%% Description:   
%%
%% Parameters:    
%%======================================================================


split_in_readable_unreadable(mnesia, ListOfTables) ->
    {ListOfTables, []};
split_in_readable_unreadable(ets, ListOfTables) ->
    Unreadable = [X || X <- ListOfTables,
		       X#tab_info.protection == private],
    {ListOfTables--Unreadable, Unreadable}.
  






create_window(KindOfTable) ->
    {MnesiaSelected, EtsSelected} = 
	case KindOfTable of
	    mnesia ->
		{true, false};
	    ets ->
		{false, true}
	end,

    gs:window(win, gs:start(), [{width, ?WINDOW_WIDTH},
				{height, ?WINDOW_HEIGHT},
				{bg, ?DEFAULT_BG_COLOR},
				{title, "Open Table"},
				{configure, true},
				{destroy, true}
			       ]),
    gs:frame(top_frame, win, [{width, ?WINDOW_WIDTH},
			      {height, 40},
			      {x, 0},
			      {y, 0},
			      {bg, ?DEFAULT_BG_COLOR},
			      {bw, 2}
			     ]),
    gs:radiobutton(mnesia_btn, top_frame, [{width, ?MIDDLE - 2},
					   {height, 40 - 4},
					   {x, 0},
					   {y, 0},
					   {bg, ?DEFAULT_BG_COLOR},
					   {fg, {0,0,0}},
					   {label, {text,"Show Mnesia tables"}},
					   {value, mnesia},
					   {align, center},
					   {select, MnesiaSelected}
					  ]),
    gs:radiobutton(ets_btn, top_frame, [{width, ?MIDDLE - 2},
					{height, 40 - 4},
					{x, ?MIDDLE - 2},
					{y, 0},
					{bg, ?DEFAULT_BG_COLOR},
					{fg, {0,0,0}},
					{label, {text,"Show ETS tables"}},
					{value, ets},
					{align, center},
					{select, EtsSelected}
				       ]),
    gs:frame(readable_frame, win, [{bg, ?DEFAULT_BG_COLOR},
				   {width, ?MIDDLE - 12},
				   {height, 250},
				   {x, ?MIDDLE + 10},
				   {y, 58}
				  ]),
    gs:frame(unreadable_frame, win, [{bg, ?DEFAULT_BG_COLOR},
				     {width, ?MIDDLE - 12},
				     {height, 250},
				     {x, ?MIDDLE + 10},
				     {y, -308}  %% Outside window!
				    ]),
    gs:label(readable_listbox_label, readable_frame, 
	     [{label, {text,"Readable tables (and owners):  "}},
	      {bg, ?DEFAULT_BG_COLOR},
	      {fg, {0, 0, 0}},
	      {align, center},
	      {width, ?MIDDLE - 25},
	      {height, 20},
	      {x, 0},
	      {y, 0}
	     ]),
    gs:label(unreadable_listbox_label, unreadable_frame, 
	     [{label, {text,"Unreadable tables (and owners):"}},
	      {bg, ?DEFAULT_BG_COLOR},
	      {fg, {0,0,0}},
	      {align, center},
	      {width, ?MIDDLE - 25},
	      {height, 20},
	      {x, 0},
	      {y, 0}
	     ]),
    gs:listbox(readable_listbox, readable_frame, 
	       [{width, ?MIDDLE - 12},
		{height, 230},
		{x, 0},
		{y, 22},
		{bg, {255,255,255}},
		{fg, {0,0,0}},
		{vscroll, right},
		{hscroll, true},
		{click, true},
		{doubleclick, true}
	       ]),
    gs:listbox(unreadable_listbox, unreadable_frame, 
	       [{width, ?MIDDLE - 12},
		{height, 230},
		{x, 0},
		{y, 22},
		{bg, {255,255,255}},
		{fg, {165,165,165}},
		{vscroll, right},
		{hscroll, true},
		{click, true}
	       ]),
    gs:checkbutton(unreadable_btn, win, 
		   [{width, ?MIDDLE - 56},
		    {height, 18},
		    {x, ?MIDDLE + 26}, 
		    {y, 322},
		    {bg, ?DEFAULT_BG_COLOR},
		    {fg, {0,0,0}},
		    {label, {text,"Show unreadable tables only"}},
		    {data, show},
		    {align, center},
		    {select, false}
		   ]),
    gs:label(node_listbox_label, win, [{label, {text,"Connected nodes:"}},
				       {bg, ?DEFAULT_BG_COLOR},
				       {fg, {0,0,0}},
				       {align, center},
				       {width, ?MIDDLE - 25},
				       {height, 20},
				       {x, 0},
				       {y, 58}
				      ]),
    gs:listbox(node_listbox, win, [{x, 0},
				   {y, 80},
				   {width, ?MIDDLE - 12},
				   {height, 230},
				   {bg, {255,255,255}},
				   {vscroll, right},
				   {hscroll, true},
				   {click, true}
				  ]),
    gs:label(info_label, win, [{label, {text, ""}},
			       {bg, ?DEFAULT_BG_COLOR},
			       {fg, {255, 0, 0}},
			       {align, center},
			       {width, ?MIDDLE - 38},
			       {height, 18},
			       {x, 0},
			       {y, 322},
			       {justify, left}
			      ]),
    BtnWidth = 70,
    gs:button(open_btn, win, [{label, {text,"Open"}},
			      {bg, ?DEFAULT_BG_COLOR},
			      {fg, {0,0,0}},
			      {align, center},
			      {width, BtnWidth},
			      {height, 30},
			      {x, (?MIDDLE - 38 - BtnWidth) / 2},
			      {y, 365}
			     ]),
    gs:button(rescan_btn, win, [{label, {text,"Rescan"}},
				{bg, ?DEFAULT_BG_COLOR},
				{fg, {0,0,0}},
				{align, center},
				{width, BtnWidth},
				{height, 30},
				{x, ?MIDDLE - BtnWidth / 2},
				{y, 365}
			       ]),
    gs:button(cancel_btn, win, [{label, {text,"Cancel"}},
				{bg, ?DEFAULT_BG_COLOR},
				{fg, {0,0,0}},
				{align, center},
				{width, BtnWidth},
				{height, 30},
				{x, ?MIDDLE + (?MIDDLE + 38) / 2 - BtnWidth / 2},
				{y, 365}
			       ]).
    







