%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%% File    : graph_placer.erl
%%% Author  : Klas Eriksson <eklas@geir>
%%% Purpose : Ported GraphLayout.java to Erlang.
%%%           Automated graph placer.
%%% Created : 29 Jul 1996 by Klas Eriksson <eklas@geir>
%%% $Id: graph_placer.erl,v 1.1 1996/11/28 08:53:47 chris Exp $
%%%

-module(graph_placer).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

-author('eklas@geir').
-revision('$Revision: 1.1 $').

-compile(export_all).
-export([relax/4]).
-export([vertex/3,vertex/4,vertex/6,edge/4,edge/7,
	 draw/1,draw/2,random_place/2]).

-include("vertex.hrl").
-include("edge.hrl").

-define(STEP,10).
-define(R2,5).


%%----------------------------------------------------------------------
%% Pre: The random:seed/3 is executed
%% Args: Graph is a digraph with the datafield as the records above.
%%       Deltamax is a float()
%% Iteration end when Deltamax or MaxIters is reached.
%% Returns: as relax/5.
%%----------------------------------------------------------------------


relax(Graph,MaxIters,Deltamax,{Xmax,Ymax}) ->
    relax(0,Graph,MaxIters,Deltamax,{Xmax,Ymax}).

%%----------------------------------------------------------------------
%% The "intelligent" part.
%% Purpose: move nodes until Deltamax is reached (most Nmax times).
%% Returns: {true|false,DeltaMax} (depending of if deltamax is reached)
%%----------------------------------------------------------------------
relax(N,G,Nmax,DeltaMax,Max) ->
    Es=digraph:edges(G),
    Vs=digraph:vertices(G),
    for1(Es,G),
    for2(Vs,Vs,G),
    case update(Vs,G,0,Max) of
	DM when DM < DeltaMax -> {true, DM};
	DM when N < Nmax -> relax(N+1,G,Nmax,DeltaMax,Max);
	DM -> {false, DM}
    end.

%%----------------------------------------------------------------------
%% Returns: <Only side effects>
%%----------------------------------------------------------------------
for1([Eid|Es],G) ->
    {E,V1,V2,D}=digraph:edge(G, Eid),
    {_,VD1}=digraph:vertex(G,V1),
    {_,VD2}=digraph:vertex(G,V2),
    #edge{len=Desired_Len}=D,
    #vertex{x=X1,y=Y1,impl={Dx1,Dy1}}=VD1,
    #vertex{x=X2,y=Y2,impl={Dx2,Dy2}}=VD2,
    Vx=X2-X1,
    Vy=Y2-Y1,
    Len=math:sqrt(lists:max([Vx*Vx+Vy*Vy,0.0001])),
    F=(Desired_Len-Len)/(Len*3),
    Dxx=F*Vx,
    Dyy=F*Vy,
    digraph:add_vertex(G,V1,VD1#vertex{impl={Dx1-Dxx,Dy1-Dyy}}),
    digraph:add_vertex(G,V2,VD2#vertex{impl={Dx2+Dxx,Dy2+Dyy}}),
    for1(Es,G);
for1([],G) -> ok.

%%----------------------------------------------------------------------
%% Returns: <Only side effects>
%%----------------------------------------------------------------------
for2([V|T],Vs,G) ->
    {Dx,Dy}=for3(Vs,G,V,0,0),
    Dlen=Dx*Dx+Dy*Dy,
    if Dlen > 0 ->
	    Dlen2=math:sqrt(Dlen)/2,
	    {_V,D} = digraph:vertex(G,V),
	    {Dxx,Dyy}=D#vertex.impl,
	    D2=D#vertex{impl={Dxx+Dx/Dlen2,Dyy+Dy/Dlen2}},
	    digraph:add_vertex(G,V,D2);
       true -> 
	    true
    end,
    for2(T,Vs,G);
for2([],Vs,G) -> ok.

%%----------------------------------------------------------------------
%% Returns: <Only side effects>
%%----------------------------------------------------------------------
for3([Vid|Vs],G,Vid,Dx,Dy) -> for3(Vs,G,Vid,Dx,Dy);
for3([Vid2|Vs],G,Vid1,Dx,Dy) ->
    {V1,D1}=digraph:vertex(G,Vid1),
    {V2,D2}=digraph:vertex(G,Vid2),
    #vertex{x=X1,y=Y1}=D1,
    #vertex{x=X2,y=Y2}=D2,
    Vx=X1-X2,
    Vy=Y1-Y2,
    case math:sqrt(Vx*Vx+Vy*Vy) of
	0.0 -> for3(Vs,G,Vid1,Dx + random:uniform(), Dy + random:uniform());
	Len when Len < 100*100 ->
	    for3(Vs,G,Vid1,Dx + Vx/Len, Dy + Vy/Len);
	Len ->
	    for3(Vs,G,Vid1,Dx,Dy)
    end;
for3([],G,Vid,Dx,Dy) -> {Dx,Dy}.

%%----------------------------------------------------------------------
%% Returns: DeltaMax
%%----------------------------------------------------------------------
update([Vid|Vs],G,DM,Max) ->
    {_,D}=digraph:vertex(G,Vid),
    #vertex{x=X,y=Y,impl={Dx,Dy},fixed=F}=D,
    if F==false ->
	    Dx2=lists:max([-?STEP,lists:min([?STEP,Dx])]),
	    Dy2=lists:max([-?STEP,lists:min([?STEP,Dy])]),
	    {Xmax,Ymax}=Max,
	    Data2=D#vertex{x=assure(X+Dx2,Xmax),
			   y=assure(Y+Dy2,Ymax),
			   impl={Dx2/2,Dy2/2}},
	    digraph:add_vertex(G,Vid,Data2),
	    update(Vs,G,lists:max([DM,Dx2,Dy2]),Max);
       true ->
	    update(Vs,G,DM,Max)
    end;
update([],S,DeltaMax,Max) -> DeltaMax.

assure(Val,Max) when Val < 0 -> 0;
assure(Val,Max) when Val > Max -> Max;
assure(Val,Max) -> Val.

%%======================================================================
%% Extra Tools:
%%======================================================================


%%----------------------------------------------------------------------
%% Returns: {A digraph vertex,GsOval}
%%----------------------------------------------------------------------

vertex(G,Canvas,{Xmax,Ymax}) ->
    vertex(G,Canvas,random:uniform()*Xmax,random:uniform()*Ymax,false,[]).
vertex(G,Canvas,X,Y) ->
    vertex(G,Canvas,X,Y,false,[]).

%% vertex/5
%%
%% Purpose: put nodeobject on the screen and store some vertexdata in a graph.
%%
%% Returns: tuple {Vertex, Node}
%%                Vertex := a DIGRAPH vertex.
%%                Node   := GS reference to a nodeobject.
%%
%% Arg: G     := a graph (digraph).
%%      Canvas:= a canvas.
%%      X,Y   := coordinates.
%%      Fixed := boolean indicating fixed or unfixed status for Name.
%%      Data  := list of items (functions/modules).
%%
vertex(G,Canvas,X,Y,Fixed, [{Module, Func_name, Arrity}]) ->
    %%CTold%%    Data = [Func_name],
    Data = [{Module, Func_name, Arrity}],
    Text_to_show = lists:append(atom_to_list(Func_name), [47 | integer_to_list(Arrity)]),
    Text = gs:text(Canvas,[{anchor, nw}, {coords,[{X,Y}]}, {fg, red}, {font, {times,12}}, {text,Text_to_show}]),
    N = #vertex{x = X,y = Y,fixed = Fixed,data = Data,gsid = Text},
    {digraph:add_vertex(G,fer_ekam(),N),Text};
vertex(G,Canvas,X,Y,Fixed,Data) ->
    [Data_atom] = Data,
    Text = gs:text(Canvas,[{anchor, nw}, {coords,[{X,Y}]}, {fg, red}, {font, {times,12}}, {text,atom_to_list(Data_atom)}]),
    N = #vertex{x = X,y = Y,fixed = Fixed,data = Data,gsid = Text},
    {digraph:add_vertex(G,fer_ekam(),N),Text}.


%%----------------------------------------------------------------------
%% Returns: {A digraph edge,GsLine}
%%----------------------------------------------------------------------

edge(G,Canvas,V1,V2) ->
    edge(G,Canvas,V1,V2,30,none,[]).
edge(G,Canvas,V1,V2,Length,Arrow,Data) ->
    {_,#vertex{x=X1,y=Y1}}=digraph:vertex(G,V1),
    {_,#vertex{x=X2,y=Y2}}=digraph:vertex(G,V2),
    Line=gs:line(Canvas,[{arrow,Arrow},lower,{fg,blue},
			 {coords,[{X1,Y1},{X2,Y2}]}]),
    E=#edge{gsid=Line,len=Length},
    {digraph:add_edge(G,fer_ekam(),V1,V2,E),Line}.

%%----------------------------------------------------------------------
%% Purpose: Random placement of the graph placer's digraph Graph
%% Returns: <Only side effects>
%%----------------------------------------------------------------------
random_place(G,{Xmax,Ymax}) ->
    lists:foreach(fun (V) ->
			  {_,Vert}=digraph:vertex(G,V),
			  Vert2=Vert#vertex{x=random:uniform()*Xmax,
					    y=random:uniform()*Ymax},
			  digraph:add_vertex(G,V,Vert2)
		  end,
		  digraph:vertices(G)).

%%----------------------------------------------------------------------
%% Purpose: Draws the graph placer's digraph Graph
%% Returns: <Only side effects>
%%----------------------------------------------------------------------
draw(G) ->
    draw(G,fun (#vertex{gsid = Text,x=X,y=Y}) ->
		   gs:config(Text,{coords,[{X,Y}]})
	   end).
		   
%%----------------------------------------------------------------------
%% Purpose: Draws the graph placer's digraph Graph
%% Usefule when: The node is user defined. 
%%               Example: A node could be an image and a label.
%% Returns: <Only side effects>
%%----------------------------------------------------------------------
draw(G,NodeUpdateFun) ->
    lists:foreach(
      fun (V) ->
	      {_,Vertex}=digraph:vertex(G,V),
	      NodeUpdateFun(Vertex),
	      update_out_edges(G,V,Vertex)
      end,
      digraph:vertices(G)).

update_out_edges(G,V,#vertex{x=X1,y=Y1}) ->
    lists:foreach(
      fun(E) ->
	      {_,_,V2,#edge{gsid=Line}}=digraph:edge(G, E),
	      {_,#vertex{x=X2,y=Y2}}=digraph:vertex(G,V2),
	      gs:config(Line,{coords,[{X1,Y1},{X2,Y2}]})
      end,
      digraph:out_edges(G,V)).

%%----------------------------------------------------------------------
%% Testing...
%%----------------------------------------------------------------------
dump(G) ->
    {lists:map(fun (V) -> digraph:vertex(G,V) end,digraph:vertices(G)),
     lists:map(fun (V) -> digraph:edge(G,V) end,digraph:edges(G))}.

reset(G) ->
    lists:foreach(fun (V) ->
			  {_,VD}=digraph:vertex(G,V),
			  digraph:add_vertex(G,V,VD#vertex{impl={0,0}})
		  end,digraph:vertices(G)).

dummy() ->
    dump(dummy),
    reset(dummy),
    dummy().

%%----------------------------------------------------------------------
%% Should be make_ref() when that otp-p1c bug is removed.
%%----------------------------------------------------------------------
fer_ekam() ->
    N=get(n),
    put(n,N+1),
    N.
