%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%% -------------------------------------------------------------------
%%% File    : xref_gp.erl
%%% Author  : Maria Holm 
%%% Purpose : To draw and interact with the graph returned from Exrefon
%%% Created : August 1996 by Maria Holm 
%%% $Id: xref_gp.erl,v 1.2 1996/12/11 10:51:24 chris Exp chris $
%%% -------------------------------------------------------------------

-module(xref_gp).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-export([create_base/3,extend/3,disp_mods/0,disp_funcs/1,ed_disp/3,
	 select_node/2,show_edges/1,destroy_edges/1,all_nodes/1,edges/1]).

-export([divide_in_mod_types/2]).

-define(Xc,300).               %% X-coord offset
-define(Yc,250).               %% Y-coord offset       
-define(R_inner,40).           %% to innermost circle
-define(R_diff, 60).           %% distance to the next level circle
-define(Y_text_distance, 12).  %% distance, on the Y-axis, between displayed text 

-include("vertex.hrl").
-include("edge.hrl").
%%CTold%%-record(vertex,{gsid,x,y,fixed,data,impl={0,0}}).
%%CTold%%-record(edge,{gsid,len,data}).

%% create_base/3
%%
%% Purpose: 1. store the selected nodes as vertices in the graph.
%%          2. put the selected modules on the screen.
%%          3. put the surounding levels of modules on the screen.
%%
%% Returns: Nothing of interest.
%%
create_base(G,Names,Level) when list(Names) ->
    %%
    %% Base    := names on functions/modules := Names.
    %% NodeIds := list of tuples containing func./mod. names and a DIGRAPH vertex.
    %%
    { Base , NodeIds } = bases(G,Names),
    %%io:format("Base length = ~w~n", [length(Base)]),
    %%io:format("NodeIds length = ~w~n", [length(NodeIds)]),
    Edges = case hd(Names) of
	     {_,_,_} ->
			{_,{Nodes,Edgs}} = exref:analyse(user_defined,
							 {exref_subgraph,
							  func_nodes_and_edges,
							  [Level,Base]}),
%%CTold%%							  [infinity,Base]}),
		    %%io:format("Nodes length = ~w~n", [length(Nodes)]),
		    %%io:format("Edgs length = ~w~n", [length(Edgs)]),
			lists:map(fun({_,A,B,_}) ->
					 {A,B}
				 end,Edgs);
	     Other ->
		 get_edges()
	 end,
    %%io:format("Edges length = ~w~n", [length(Edges)]),
    place_graph(G,Level,Edges,Base,NodeIds,1).

%% bases/2   (Block of functions)
%%
%% Purpose: places the selected nodeobjects on the screen mapped onto a circle
%%          and insert them as verices in a graph, along with a certain datarecord. 
%%          
%% Returns: tuple {Base, NodeIds}
%%                Base    := names on functions/modules := Names.
%%                NodeIds := list of tuples containing func./mod.
%%                           names and a corresponding DIGRAPH vertex.
%%                 
%% Arg:     G     := a graph (digraph).
%%          Names := list of functions/modules.
%%
bases(G,Names) -> 
    Length = length(Names),
    All_coords = generate_all_coords(?Y_text_distance, Length, ?R_inner),
    AddAngle = (2*math:pi())/Length, %% Angle difference between Names placed on a circle.
    bases(G,Names,[],[],{?Xc,?Yc,30,0,AddAngle}, All_coords).
%%CTold%%    bases(G,Names,[],[],{?Xc,?Yc,30,0,AddAngle}).

bases(G,[],Base,NodeIds,Dim, _) -> {Base,NodeIds};
bases(G,[Name|Rest],Base,NodeIds,Dim, [{X, Y} | Coords]) ->
    %%CTold%%{X,Y} = get_coords(Dim), %% Coordinates belonging to Name.
    Vertex = vertex(G, X, Y, true, Name), %% a DIGRAPH vertex.
    bases(G,Rest,[Name|Base],[{Name,Vertex}|NodeIds],get_next_dimension(Dim), Coords).
%%
%% bases/2  (End of block)


%% generate_all_coords/3
%%
%% Purpose: generates at least so many coordinates as there are items (N_items) to display.
%%          The coordinates are taken from a circle with radius Radius, and
%%          where the Y-coordinates have a konstant difference (distance). The
%%          later issue is to prevent texts to be displayed upon each other.
%% Returns: [{X1,Y1}, {X2,Y2}, ...]
%%
generate_all_coords( Y_dist , N_items , Radius ) ->
    R_parts = (N_items + 3) div 4, %% nearest higher (or equal) value == 0 (mod 4)
    R = lists:max([Radius, Y_dist * R_parts]), %% assert Radius >= N_items*Y_dist
    New_Y_dist = R div R_parts,    %% if a new radius, justify the text-Y-difference
    R_square = R * R,
    Some_R_parts = generate_parts_of_R(0, New_Y_dist, R_parts - 1), %% head should be nearest the center of the circle
    Coords = fun ( Y , Acc_coords) -> %% add the 2 coords on the circle for a cetain Y.
		     X = trunc(math:sqrt(R_square - Y*Y) + 0.5),   %% just pythagoras theorem
		     [{-X + ?Xc,  Y + ?Yc},
		      { X + ?Xc,  Y + ?Yc},
		      {-X + ?Xc, -Y + ?Yc},
		      { X + ?Xc, -Y + ?Yc}   | Acc_coords]
	      end,
    Start_coords = [{-R + ?Xc, 0 + ?Yc}, {R + ?Xc, 0 + ?Yc}],  %% on the x-axis
    All_but_2    = lists:foldl(Coords, Start_coords, Some_R_parts),
    All_coords   = [{0 + ?Xc, R + ?Yc}, {0 + ?Xc, -R + ?Yc} | All_but_2],              %% on the y-axis
    %%io:format("Coords = ~w~n",[All_coords]),
    All_coords.

%% generate_parts_of_R/3
%%
%% Purpose: generate a list with elements showing an accumulated sum.
%%
%% Returns: [] | [Start_value + 1*Incr, Start_value + 2*Incr, ... ]
%%
generate_parts_of_R(Start_value, Incr, N_times) when N_times > 0 ->
    gen_R_parts(Start_value, Incr, N_times);
generate_parts_of_R(_, _, _) -> [].
gen_R_parts(Sum, Incr, 1) -> [Sum + Incr];
gen_R_parts(Sum, Incr, N_times) ->
    [Incr + Sum | gen_R_parts(Sum + Incr, Incr, N_times - 1) ].


%% vertex/5
%%
%% Purpose: creates a vertex by calling graph_placer
%%          returns the id of the digraph_vertex.
%%
%% Returns: a DIGRAPH vertex.
%%
%% Arg: G     := a graph (digraph).
%%      X,Y   := coordinates.
%%      Fixed := boolean indicating fixed or unfixed status for Name.
%%      Name  := name on item (function/module).
%%
%%CTold%%
vertex(G,X,Y,Fixed,Name) ->
    %%
    %% Vertex := DIGRAPH.
    %% Node   := GS reference.
    %%
    {Vertex,Node} = graph_placer:vertex(G,canvas,X,Y,Fixed,[Name]),
    gs:config(Node,[{buttonpress,true},{buttonrelease,true},{enter,true}]),
    gs:config(Node,{data,{digraph_vertex,Vertex}}),
    case Fixed of 
	true -> gs:config(Node,{fg,yellow});
	false -> gs:config(Node,{fg,red})
    end,
    Vertex;
%%CTend%%

vertex(G,X,Y,Fixed,Name) ->
    {Vertex,Node} = graph_placer:vertex(G,canvas,X,Y,Fixed,[Name]),
    gs:config(Node,[{buttonpress,true},{buttonrelease,true},{enter,true}]),
    gs:config(Node,{data,{digraph_vertex,Vertex}}),
    case Fixed of 
	true -> gs:config(Node,{fill,yellow});
	false -> gs:config(Node,{fill,red})
    end,
    Vertex.


%% place_graph/6
%%
%% Purpose: put the surounding levels of modules (to the selected ones)
%%          on the screen, and insert them as verices in a graph along
%%          with a certain datarecord.
%% Returns: [{Name, Vertex}, ...]
%%          Name   = module/function name
%%          Vertex = DIGRAPH vertex
%%
place_graph(G,0,Edges,Front,NodeIds,R_level) -> NodeIds;
place_graph(G,Level,Edges,Front,NodeIds,R_level) ->
    NewFront = next_level(Edges,Front,NodeIds,[]),
    %%io:format("PLACE_GRAPH: NewFront length = ~w~n", [length(NewFront)]),
    case length(NewFront) of
	0 ->
	    NodeIds;
	Length ->
	    Angle = (2*math:pi())/Length,   %% not in use any longer
	    NewDim = {?Xc,?Yc,?R_inner + ?R_diff*R_level,0,Angle},   %% not in use any longer
	    All_coords = generate_all_coords(?Y_text_distance, Length , ?R_inner + ?R_diff * R_level ),
	    NewIds = make_level(G,NewFront,NewDim,NodeIds, All_coords),
	    place_graph(G,dec(Level),Edges,NewFront,NewIds,R_level+1)
    end.

dec(infinity) -> infinity;
dec(Int) -> Int - 1.

%% next_level/4
%% Finds the next level

%% next_level/4
%%
next_level([],Front,NodeIds,Acc) -> Acc;   
next_level([{A,B}|Edges],Front,NodeIds,Acc) ->
    AF = lists:member(A,Front),
    BN = lists:keysearch(B,1,NodeIds),
    BA = lists:member(B,Acc),
    case {AF,BN,BA} of
	{true,false,false} ->
	    next_level(Edges,Front,NodeIds,[B|Acc]);
	_ ->
	    next_level(Edges,Front,NodeIds,Acc)
    end.
%%
%%next_level([],Front,NodeIds,Acc) -> Acc;   
%%next_level([{A,B}|Edges],Front,NodeIds,Acc) ->
%%    AF = lists:member(A,Front),
%%    BF = lists:member(B,Front),
%%    AI = lists:keysearch(A,1,NodeIds),
%%    BI = lists:keysearch(B,1,NodeIds),
%%    ANo = lists:member(A,Acc),
%%    BNo = lists:member(B,Acc),
%%    case {AF,BF,AI,BI,ANo,BNo} of
%%	{true,_,  _,false,  _,false} -> 
%%	    next_level(Edges,Front,NodeIds,[B|Acc]);
%%	{_,true,  false,_,  false,_} ->
%%	    next_level(Edges,Front,NodeIds,[A|Acc]);					     
%%	Other ->
%%	    next_level(Edges,Front,NodeIds,Acc)
%%    end.

%% make_level/4
%% creates a level and finds the coordinates for its nodes

make_level(G,[],Dim,NodeIds, _) -> NodeIds;
make_level(G,[Name|Rest],Dim,NodeIds, [{ X , Y } | Coords]) ->
    case lists:keysearch(Name,1,NodeIds) of
	false ->
%%CTold%%	    {X,Y} = get_coords(Dim),
	    Vertex = vertex( G, X, Y, false, Name),
	    make_level(G,Rest,get_next_dimension(Dim),[{Name,Vertex}|NodeIds], Coords);
	{value,_} ->
	    make_level(G,Rest,Dim,NodeIds, [{ X , Y } | Coords])
    end.

%% create_edges/3
%% creates edges if both callingfunction and calledfunction are in the graph

create_edges(Graph,[],NodeIds) -> Graph;
create_edges(Graph,[{CallingF,CalledF}|Edges],NodeIds) ->
    F1 = lists:keysearch(CallingF,1,NodeIds),
    F2 = lists:keysearch(CalledF,1,NodeIds),
    case {F1,F2} of
	{{value,{_,Id1}},{value,{_,Id2}}} ->
	    {_,Edge} = graph_placer:edge(Graph,canvas,Id1,Id2,70,last,[]),
	    create_edges(Graph,Edges,NodeIds);
	Other ->
	    create_edges(Graph,Edges,NodeIds)
    end.

%% extend/3
%% extends the vertices in the list Vertices
%% Direction can be:
%%  in - extends the graph with functions calling to Vertices, creates edges
%%  out - extends the graph with functions called to by Vertices, creates edges
%%  all - extends the graph with fuctions calling to and called by Vertices
 
extend(G,Vertices,Direction) ->
    NodeIds = lists:map(fun (Id) ->
			       {_,D} = digraph:vertex(G,Id),
			       Name = hd(D#vertex.data),
			       {Name,Id}
		       end,digraph:vertices(G)),
    {Names,Type,Rad} = get_names(G,Vertices),
    E = case Type of
	    {_,_,_} ->
		{_,ExrEdg} = exref:analyse(user_defined,
					   {xref_gp,edges}), %% --> 10
		ExrEdg;
	    Other -> get_edges()
	end,
    {Nodes,Edges} = peel(G,E,Names,NodeIds,Direction), 
    Ids = case length(Nodes) of
	      0 ->
		  NodeIds;
	      Length ->
		  %%io:format("EXTEND: Ids length = ~w~n", [Length]),
		  AddAngle = (2*math:pi())/Length,  %% no longer in use
		  %%CTold%%Radius = Rad+70,                  
		  Radius = trunc(Rad + ?R_diff + 0.5),
		  All_coords = generate_all_coords(?Y_text_distance, Length, Radius),
		  add_nodes(G,Nodes,{?Xc,?Yc,Radius,0,AddAngle},NodeIds, All_coords)
	  end,
    create_edges(G,Edges,Ids).

%% peel/5,7
%% peels of the edges not to be displayed and adds the nodes to be created.

peel(G,Edges,Names,Ids,Direction) -> peel(G,Edges,Names,Ids,[],[],Direction).
peel(G,[],Names,Ids,Nodes,Edges,Direction) -> {Nodes,Edges};
peel(G,[{A,B}|R],Names,Ids,Nodes,Edges,in) ->
    BN = lists:member(B,Names),
    AI = lists:keysearch(A,1,Ids),
    ANo = lists:member(A,Nodes),
    case {BN,AI,ANo} of
	{true,false,false} -> 
	    peel(G,R,Names,Ids,[A|Nodes],[{A,B}|Edges],in);
	{true,_,_} -> peel(G,R,Names,Ids,Nodes,[{A,B}|Edges],in);
	Other -> peel(G,R,Names,Ids,Nodes,Edges,in)
    end;
peel(G,[{A,B}|R],Names,Ids,Nodes,Edges,out) ->
    AN = lists:member(A,Names),
    BI = lists:keysearch(B,1,Ids),
    BNo = lists:member(B,Nodes),
    case {AN,BI,BNo} of
	{true,false,false} -> 
	    peel(G,R,Names,Ids,[B|Nodes],[{A,B}|Edges],out);
	{true,_,_} -> peel(G,R,Names,Ids,Nodes,[{A,B}|Edges],out);
	Other -> peel(G,R,Names,Ids,Nodes,Edges,out)
    end;
peel(G,[{A,B}|R],Names,Ids,Nodes,Edges,all) ->
    AN = lists:member(A,Names),
    BN = lists:member(B,Names),
    AI = lists:keysearch(A,1,Ids),
    BI = lists:keysearch(B,1,Ids),
    ANo = lists:member(A,Nodes),
    BNo = lists:member(B,Nodes),
    case {AN,BN,AI,BI,ANo,BNo} of
	{true,_,  _,false,  _,false} -> 
	    peel(G,R,Names,Ids,[B|Nodes],Edges,all);
	{_,true,  false,_,  false,_} ->
	    peel(G,R,Names,Ids,[A|Nodes],Edges,all);
	Other -> peel(G,R,Names,Ids,Nodes,Edges,all)
    end.

add_nodes(G,[],Dim,NodeIds, Coords) -> NodeIds;
add_nodes(G,[Name|Rest],Dim,NodeIds, [{X,Y}|Coords]) ->
%%CTold%%    {X,Y} = get_coords(Dim),
    Vertex = vertex(G,X,Y,false,Name),
    add_nodes(G,Rest,get_next_dimension(Dim),[{Name,Vertex}|NodeIds], Coords).

%% get_named/2,5
%% extracts the names of Vertices,
%% returns Names,type of graph (if it's modules or funtions) and
%% radius the nodes should be displayed on.

get_names(G,Vertices) -> get_names(G,Vertices,[],type,0).
get_names(G,[],Names,Type,Radius) -> {Names,Type,Radius};
get_names(G,[Id|Rest],Names,Type,Radius) ->
    {_,D} = digraph:vertex(G,Id),
    #vertex{data=[Name],x=X,y=Y} = D,
    R = math:sqrt(((X-?Xc)*(X-?Xc))+((Y-?Yc)*(Y-?Yc))),
    get_names(G,Rest,[Name|Names],Name,max(R,Radius)).


max(A,B) when A>B -> A;
max(A,B) when A<B -> B;
max(A,B) -> A.

get_next_dimension({Xc,Yc,Radius,Angle,Addition}) ->
    {Xc,Yc,Radius,Angle+Addition,Addition}.

get_coords({Xc,Yc,Radius,Angle,Addition}) ->
    {Xc+Radius*math:cos(Angle), Yc+Radius*math:sin(Angle)}.

%% get_edges/0
%% returnes all module edges

get_edges() -> 
    All = element(2,element(3,exref:analyse(module_call))).

add([],List) -> List;
add([E|Rest],List) ->
    case lists:member(E,List) of
	true ->
	    add(Rest,List);
	false ->
	    add(Rest,[E|List])
    end.

%% disp_mods/0
%%
%% Purpose: displayed and return a sorted list of loaded modules,
%%          library modules and  undefined modules names.
%%
%%
disp_mods() ->
    Undef = exref:analyse(undefined_functions), %% -> {undefined_functions, get(modules), undef2(Vs1, S)};
    SMods = lists:sort(element(2,Undef)),
    UMods = mark_undef(element(3,Undef)),
    LibMods = mark_lib(SMods,UMods),
    Mods = lists:append([SMods, LibMods, UMods]),
    gs:config(modlb,{items,Mods}).

divide_in_mod_types(Modules, Undefined_modules) ->
    Def_mod   = lists:sort(Modules),
    Undef_mod = mark_undef(Undefined_modules),
    Lib_mod   = mark_lib(Def_mod, Undef_mod),
    {Def_mod, Lib_mod, Undef_mod}.


mark_undef(UList) -> mark_undef(UList,[]).
mark_undef([],UMods) -> lists:sort(UMods);
mark_undef([First|Rest],UMods) ->
    F = element(2,First),
    M = element(1,F),
    NewUMods = ordsets:add_element({undefined,M},UMods),
    mark_undef(Rest,NewUMods).

mark_lib(SMods,UMods) -> 
    Edges = get_edges(),
    mark_lib(Edges,SMods,UMods,[]).
mark_lib([],SMods,UMods,LibMods) -> 
    lists:sort(LibMods);
mark_lib([{M1,M2}|Edges],SMods,UMods,LibMods) ->
    M1S = lists:member(M1,SMods),
    M1U = lists:keysearch(M1,2,UMods),
    M1L = lists:keysearch(M1,2,LibMods),
    M2S = lists:member(M2,SMods),
    M2U = lists:keysearch(M2,2,UMods),
    M2L = lists:keysearch(M2,2,LibMods),
    NewLib = case {M1S,M1U,M1L} of
		 {false,false,false} -> [{'OTP',M1}|LibMods];
		 _ -> LibMods
	     end,
    case {M2S,M2U,M2L} of
	{false,false,false} -> 
	    mark_lib(Edges,SMods,UMods,[{'OTP',M2}|NewLib]);
	_ -> mark_lib(Edges,SMods,UMods,NewLib)
    end.

%% disp_funcs/1
%% displayed all functions in the exref graph belonging to the module Txt

disp_funcs(Txt) ->
    case Txt of
	[] ->
	    ok;
	Other ->
	    String = lists:append(Txt,". "),
	    {ok,Ts,_} = erl_scan:string(String),
	    {ok,Mod} = erl_parse:parse_term(Ts),
	    Funcs = case Mod of 
			{'OTP',Module} ->
			    {_,Ex} = exref:analyse(user_defined,
						   {xref_gp,all_nodes}),
			    extract_lib(Module,Ex);
			{undefined,Module} ->
			    Ex = exref:analyse(undefined_functions),
			    extract_undef(Module,Ex);
			Module ->
			    Ex = exref:analyse(call,Mod),
			    extract_funcs(Ex)
		    end,
	    gs:config(funclb,{items,Funcs})
    end.

extract_lib(Module,Ex) -> extract_lib(Module,Ex,[]).
extract_lib(Module,[],Funcs) -> lists:sort(Funcs);
extract_lib(Module,[First|Rest],Funcs) ->
    case element(1,First) of
	Module ->
	    extract_lib(Module,Rest,[First|Funcs]);
	Other ->
	    extract_lib(Module,Rest,Funcs)
    end.

extract_undef(Module,Ex) -> extract_undef(Module,element(3,Ex),[]).
extract_undef(Module,[],Funcs) -> lists:sort(Funcs);
extract_undef(Module,[First|Rest],Funcs) ->
    F = element(2,First),
    case element(1,F) of
	Module ->
	    extract_undef(Module,Rest,[F|Funcs]);
	Ohter ->
	    extract_undef(Module,Rest,Funcs)
    end.

extract_funcs(Ex) -> extract_funcs(element(3,Ex),[]).
extract_funcs([],Funcs) -> lists:sort(Funcs);
extract_funcs([First|Rest],Funcs) ->
    L = element(3,First),
    F = {lists:nth(3,L),lists:nth(4,L),lists:nth(5,L)},
    extract_funcs(Rest,[F|Funcs]).
	 
%% ed_disp/3
%% calls exref:analyse(Type) and displayes the result in an editor in a window

ed_disp(Exref,Type,Owner) ->
    process_flag(trap_exit, true),
    S = gs:start(),
    Win = gs:create(window,S,[{width,300},{height,300},{title,Type}]),
    E = gs:create(editor,Win,[{x,0},{y,0},{width,300},{height,250},
			  {vscroll,right}]),
    C = gs:create(button,Win,[{x,20},{y,260},{width,120},
			      {label,{text,"Close"}}]),
    gs:config(Win,{leave,true}),
    case Exref of
	{error, Type, _List} ->
	    gs:config(E,{insert,{insert,Exref}});
	{Type, _Relatet2mods, Mods } ->
	    ed_list(Mods, E);
	{Type, Rec_mods } ->
	    ed_list(Rec_mods, E)
    end,
    gs:config(Win,{map,true}),
    ed_loop(Win,C,Owner).

ed_list([],E) -> ok;
ed_list([{Type,Tuple}|Rest],E) when tuple(Tuple) ->
    File = {element(3,Tuple),element(4,Tuple),element(5,Tuple)},
    Test = gs:config(E,[{insert,{insert,File}},
			{insert,{insert,"\n"}}]),
    ed_list(Rest,E);
ed_list([[Type, Rec_mods ]|Rest],E) when list(Rec_mods) ->  %% C.T.: stupid structure from EXREF.ERL !
    display(E, Type, Rec_mods),
    ed_list(Rest,E).

display(E, Start_mod, Rec_mods) ->
    gs:config(E, [{insert, {insert, "{" }}]),
    gs:config(E, [{insert, {insert, Start_mod}}]),
    gs:config(E, [{insert, {insert, "\n  [" }}]),
    disp_rec_mods(E, Rec_mods).

disp_rec_mods(E, [H|[]]) ->
    gs:config(E, [{insert, {insert, H}}]),
    gs:config(E, [{insert, {insert, "] }\n"}}]);
disp_rec_mods(E, [H|T]) ->
    gs:config(E, [{insert, {insert, H}}]),
    gs:config(E, [{insert, {insert, ",\n   "}}]),
    disp_rec_mods(E, T).



ed_loop(Win,C,Owner) ->
    receive
	{'EXIT', Owner, Reason} ->
	    exit(normal);
	{gs,Win,leave,_,_} ->
	    Owner ! {ed_loop,leave},
	    ed_loop(Win,C,Owner);
	{gs,C,click,_,_} ->
	    gs:destroy(Win);
	Other ->
	    ed_loop(Win,C,Owner)
    end.

%% select_node/2
%% finds and selects the node Txt

select_node(G,Txt) ->
    String = lists:append(Txt,". "),
    {ok,Ts,_} = erl_scan:string(String),
    {ok,Selected} = erl_parse:parse_term(Ts),
    Vertices = digraph:vertices(G),
    All = lists:map(fun (Id) ->
			   {_,V} = digraph:vertex(G,Id),
			   #vertex{gsid=N,fixed=F,data=[Name]} = V,
			   {Name,N,F,Id}
		   end,Vertices),
    case lists:keysearch(Selected,1,All) of
	{value,{Selected,Node,Fixed,Gid}} ->
	    gs:config(Node,{fill,green}),
	    gs:config(nodeentry,{text,Selected}),
	    gs:config(fix,{select,Fixed}),
	    {Node,Gid,Fixed,Selected};
	false ->
	    gs:config(nodeentry,{text,"No such node"})
    end.

%% show_edges/1
%% shows all the edges in the graph

show_edges(G) -> 
    Vertices = digraph:vertices(G),
    V1 = hd(Vertices),
    {_,D} = digraph:vertex(G,V1),
    #vertex{data=[Name]} = D,
    AllEdges = case Name of
		   {_,_,_} ->
		       {_,ExrEdg} = exref:analyse(user_defined,
						  {xref_gp,edges}),
		       ExrEdg;
		   Other -> get_edges()
	       end,	    
    NodeIds = lists:map(fun (Id) ->
			       {_,Ver} = digraph:vertex(G,Id),
			       N = hd(Ver#vertex.data),
			       {N,Id}
		       end,Vertices),
    create_edges(G,AllEdges,NodeIds).

%% destroy_edges/1
%% destroys all the edges in the graph

destroy_edges(G) ->
    lists:foreach(fun(Eid) ->
			 {_,_,_,E} = digraph:edge(G,Eid),
			 GsId = E#edge.gsid,
			 gs:destroy(GsId),
			 digraph:del_edge(G,Eid)
		 end,digraph:edges(G)).


%% all_nodes/1
%% returns all the vertices in the exref digraph

all_nodes(Exg) -> digraph:vertices(Exg).
%% edges/1
%% returns all the edges in the exref digraph as tuples of 
%% {callingfunction,calledfunction} 

edges(Exg) -> 
    lists:map(fun (Eid) ->
		     {_,A,B,_} = digraph:edge(Exg,Eid),
		     {A,B}
	     end,digraph:edges(Exg)).





